# Unix SMB/CIFS implementation.
# Copyright (C) Sean Dague <sdague@linux.vnet.ibm.com> 2011
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import time
import base64
import ldb
from samba.tests.samba_tool.base import SambaToolCmdTest
from samba import (
        credentials,
        nttime2unix,
        dsdb,
        werror,
        )
from samba.ndr import ndr_unpack
from samba.dcerpc import drsblobs
from samba.common import get_bytes
from samba.common import get_string
from samba.tests import env_loadparm


class UserCmdTestCase(SambaToolCmdTest):
    """Tests for samba-tool user subcommands"""
    users = []
    samdb = None

    def setUp(self):
        super().setUp()
        self.samdb = self.getSamDB("-H", "ldap://%s" % os.environ["DC_SERVER"],
                                   "-U%s%%%s" % (os.environ["DC_USERNAME"], os.environ["DC_PASSWORD"]))

        # Modify the default template homedir
        lp = self.get_loadparm()
        self.template_homedir = lp.get('template homedir')
        lp.set('template homedir', '/home/test/%D/%U')

        self.users = []
        self.users.append(self._randomUser({"name": "sambatool1", "company": "comp1"}))
        self.users.append(self._randomUser({"name": "sambatool2", "company": "comp1"}))
        self.users.append(self._randomUser({"name": "sambatool3", "company": "comp2"}))
        self.users.append(self._randomUser({"name": "sambatool4", "company": "comp2"}))
        self.users.append(self._randomPosixUser({"name": "posixuser1"}))
        self.users.append(self._randomPosixUser({"name": "posixuser2"}))
        self.users.append(self._randomPosixUser({"name": "posixuser3"}))
        self.users.append(self._randomPosixUser({"name": "posixuser4"}))
        self.users.append(self._randomUnixUser({"name": "unixuser1"}))
        self.users.append(self._randomUnixUser({"name": "unixuser2"}))
        self.users.append(self._randomUnixUser({"name": "unixuser3"}))
        self.users.append(self._randomUnixUser({"name": "unixuser4"}))

        # Make sure users don't exist
        for user in self.users:
            if self._find_user(user["name"]):
                self.runsubcmd("user", "delete", user["name"])

        # setup the 12 users and ensure they are correct
        for user in self.users:
            (result, out, err) = user["createUserFn"](user)

            self.assertCmdSuccess(result, out, err)
            self.assertEqual(err, "", "Shouldn't be any error messages")
            if 'unix' in user["name"]:
                self.assertIn("Modified User '%s' successfully" % user["name"],
                              out)
            else:
                self.assertIn("User '%s' added successfully" % user["name"],
                              out)

            user["checkUserFn"](user)

    def tearDown(self):
        super().tearDown()
        # clean up all the left over users, just in case
        for user in self.users:
            if self._find_user(user["name"]):
                self.runsubcmd("user", "delete", user["name"])
        lp = env_loadparm()
        # second run of this test
        # the cache is still there and '--cache-ldb-initialize'
        # will fail
        cachedb = lp.private_path("user-syncpasswords-cache.ldb")
        if os.path.exists(cachedb):
            os.remove(cachedb)
        lp.set('template homedir', self.template_homedir)

    def test_newuser(self):
        # try to add all the users again, this should fail
        for user in self.users:
            (result, out, err) = self._create_user(user)
            self.assertCmdFail(result, "Ensure that create user fails")
            self.assertIn("LDAP error 68 LDAP_ENTRY_ALREADY_EXISTS", err)

        # try to delete all the 4 users we just added
        for user in self.users:
            (result, out, err) = self.runsubcmd("user", "delete", user["name"])
            self.assertCmdSuccess(result, out, err, "Can we delete users")
            found = self._find_user(user["name"])
            self.assertIsNone(found)

        # test adding users with --use-username-as-cn
        for user in self.users:
            (result, out, err) = self.runsubcmd("user", "create", user["name"], user["password"],
                                                "--use-username-as-cn",
                                                "--surname=%s" % user["surname"],
                                                "--given-name=%s" % user["given-name"],
                                                "--job-title=%s" % user["job-title"],
                                                "--department=%s" % user["department"],
                                                "--description=%s" % user["description"],
                                                "--company=%s" % user["company"],
                                                "-H", "ldap://%s" % os.environ["DC_SERVER"],
                                                "-U%s%%%s" % (os.environ["DC_USERNAME"], os.environ["DC_PASSWORD"]))

            self.assertCmdSuccess(result, out, err)
            self.assertEqual(err, "", "Shouldn't be any error messages")
            self.assertIn("User '%s' added successfully" % user["name"], out)

            found = self._find_user(user["name"])

            self.assertEqual("%s" % found.get("cn"), "%(name)s" % user)
            self.assertEqual("%s" % found.get("name"), "%(name)s" % user)

    def test_newuser_weak_password(self):
        # Ensure that when we try to create a user over LDAP (thus no
        # transactions) and the password is too weak, we do not get a
        # half-created account.

        def cleanup_user(username):
            try:
                self.samdb.deleteuser(username)
            except Exception as err:
                estr = err.args[0]
                if 'Unable to find user' not in estr:
                    raise

        server = os.environ['DC_SERVER']
        dc_username = os.environ['DC_USERNAME']
        dc_password = os.environ['DC_PASSWORD']

        username = self.randomName()
        password = 'a'

        self.addCleanup(cleanup_user, username)

        # Try to add the user and ensure it fails.
        result, out, err = self.runsubcmd('user', 'add',
                                          username, password,
                                          '-H', f'ldap://{server}',
                                          f'-U{dc_username}%{dc_password}')
        self.assertCmdFail(result)
        self.assertIn('Failed to add user', err)
        self.assertIn('LDAP_UNWILLING_TO_PERFORM', err)
        self.assertIn(f'{werror.WERR_PASSWORD_RESTRICTION:08X}', err)

        # Now search for the user, and make sure we don't find anything.
        res = self.samdb.search(self.samdb.domain_dn(),
                                expression=f'(sAMAccountName={username})',
                                scope=ldb.SCOPE_SUBTREE)
        self.assertEqual(0, len(res), 'expected not to find the user')

    def _verify_supplementalCredentials(self, ldif,
                                        min_packages=3,
                                        max_packages=6):
        msgs = self.samdb.parse_ldif(ldif)
        (changetype, obj) = next(msgs)

        self.assertIn("supplementalCredentials", obj, "supplementalCredentials attribute required")
        sc_blob = obj["supplementalCredentials"][0]
        sc = ndr_unpack(drsblobs.supplementalCredentialsBlob, sc_blob)

        self.assertGreaterEqual(sc.sub.num_packages,
                                min_packages, "min_packages check")
        self.assertLessEqual(sc.sub.num_packages,
                             max_packages, "max_packages check")

        if max_packages == 0:
            return

        def find_package(packages, name, start_idx=0):
            for i in range(start_idx, len(packages)):
                if packages[i].name == name:
                    return (i, packages[i])
            return (None, None)

        # The ordering is this
        #
        # Primary:Kerberos-Newer-Keys (optional)
        # Primary:Kerberos
        # Primary:WDigest
        # Primary:CLEARTEXT (optional)
        # Primary:SambaGPG (optional)
        #
        # And the 'Packages' package is insert before the last
        # other package.

        nidx = 0
        (pidx, pp) = find_package(sc.sub.packages, "Packages", start_idx=nidx)
        self.assertIsNotNone(pp, "Packages required")
        self.assertEqual(pidx + 1, sc.sub.num_packages - 1,
                         "Packages needs to be at num_packages - 1")

        (knidx, knp) = find_package(sc.sub.packages, "Primary:Kerberos-Newer-Keys",
                                    start_idx=nidx)
        if knidx is not None:
            self.assertEqual(knidx, nidx, "Primary:Kerberos-Newer-Keys at wrong position")
            nidx = nidx + 1
            if nidx == pidx:
                nidx = nidx + 1

        (kidx, kp) = find_package(sc.sub.packages, "Primary:Kerberos",
                                  start_idx=nidx)
        self.assertIsNotNone(pp, "Primary:Kerberos required")
        self.assertEqual(kidx, nidx, "Primary:Kerberos at wrong position")
        nidx = nidx + 1
        if nidx == pidx:
            nidx = nidx + 1

        (widx, wp) = find_package(sc.sub.packages, "Primary:WDigest",
                                  start_idx=nidx)
        self.assertIsNotNone(pp, "Primary:WDigest required")
        self.assertEqual(widx, nidx, "Primary:WDigest at wrong position")
        nidx = nidx + 1
        if nidx == pidx:
            nidx = nidx + 1

        (cidx, cp) = find_package(sc.sub.packages, "Primary:CLEARTEXT",
                                  start_idx=nidx)
        if cidx is not None:
            self.assertEqual(cidx, nidx, "Primary:CLEARTEXT at wrong position")
            nidx = nidx + 1
            if nidx == pidx:
                nidx = nidx + 1

        (gidx, gp) = find_package(sc.sub.packages, "Primary:SambaGPG",
                                  start_idx=nidx)
        if gidx is not None:
            self.assertEqual(gidx, nidx, "Primary:SambaGPG at wrong position")
            nidx = nidx + 1
            if nidx == pidx:
                nidx = nidx + 1

        self.assertEqual(nidx, sc.sub.num_packages, "Unknown packages found")

    def test_setpassword(self):
        expect_nt_hash = bool(int(os.environ.get("EXPECT_NT_HASH", "1")))

        for user in self.users:
            newpasswd = self.random_password(16)
            (result, out, err) = self.runsubcmd("user", "setpassword",
                                                user["name"],
                                                "--newpassword=%s" % newpasswd,
                                                "-H", "ldap://%s" % os.environ["DC_SERVER"],
                                                "-U%s%%%s" % (os.environ["DC_USERNAME"], os.environ["DC_PASSWORD"]))
            self.assertCmdSuccess(result, out, err, "Ensure setpassword runs")
            self.assertEqual(err, "", "setpassword with url")
            self.assertMatch(out, "Changed password OK", "setpassword with url")

        attributes = "sAMAccountName,unicodePwd,supplementalCredentials,virtualClearTextUTF8,virtualClearTextUTF16,virtualSSHA,virtualSambaGPG"
        (result, out, err) = self.runsubcmd("user", "syncpasswords",
                                            "--cache-ldb-initialize",
                                            "--attributes=%s" % attributes,
                                            "--decrypt-samba-gpg")
        self.assertCmdSuccess(result, out, err, "Ensure syncpasswords --cache-ldb-initialize runs")
        self.assertEqual(err, "", "getpassword without url")
        cache_attrs = {
            "objectClass": {"value": "userSyncPasswords"},
            "samdbUrl": {},
            "dirsyncFilter": {},
            "dirsyncAttribute": {},
            "dirsyncControl": {"value": "dirsync:1:0:0"},
            "passwordAttribute": {},
            "decryptSambaGPG": {},
            "currentTime": {},
        }
        for a in cache_attrs.keys():
            v = cache_attrs[a].get("value", "")
            self.assertMatch(out, "%s: %s" % (a, v),
                             "syncpasswords --cache-ldb-initialize: %s: %s out[%s]" % (a, v, out))

        (result, out, err) = self.runsubcmd("user", "syncpasswords", "--no-wait")
        self.assertCmdSuccess(result, out, err, "Ensure syncpasswords --no-wait runs")
        self.assertEqual(err, "", "syncpasswords --no-wait")
        self.assertMatch(out, "dirsync_loop(): results 0",
                         "syncpasswords --no-wait: 'dirsync_loop(): results 0': out[%s]" % (out))
        for user in self.users:
            self.assertMatch(out, "sAMAccountName: %s" % (user["name"]),
                             "syncpasswords --no-wait: 'sAMAccountName': %s out[%s]" % (user["name"], out))

        for user in self.users:
            newpasswd = self.random_password(16)
            creds = credentials.Credentials()
            creds.set_anonymous()
            creds.set_password(newpasswd)
            unicodePwd = base64.b64encode(creds.get_nt_hash()).decode('utf8')
            virtualClearTextUTF8 = base64.b64encode(get_bytes(newpasswd)).decode('utf8')
            virtualClearTextUTF16 = base64.b64encode(get_string(newpasswd).encode('utf-16-le')).decode('utf8')

            (result, out, err) = self.runsubcmd("user", "setpassword",
                                                user["name"],
                                                "--newpassword=%s" % newpasswd)
            self.assertCmdSuccess(result, out, err, "Ensure setpassword runs")
            self.assertEqual(err, "", "setpassword without url")
            self.assertMatch(out, "Changed password OK", "setpassword without url")

            (result, out, err) = self.runsubcmd("user", "syncpasswords", "--no-wait")
            self.assertCmdSuccess(result, out, err, "Ensure syncpasswords --no-wait runs")
            self.assertEqual(err, "", "syncpasswords --no-wait")
            self.assertMatch(out, "dirsync_loop(): results 0",
                             "syncpasswords --no-wait: 'dirsync_loop(): results 0': out[%s]" % (out))
            self.assertMatch(out, "sAMAccountName: %s" % (user["name"]),
                             "syncpasswords --no-wait: 'sAMAccountName': %s out[%s]" % (user["name"], out))
            self.assertMatch(out, "# unicodePwd::: REDACTED SECRET ATTRIBUTE",
                             "getpassword '# unicodePwd::: REDACTED SECRET ATTRIBUTE': out[%s]" % out)
            if expect_nt_hash or "virtualSambaGPG:: " in out:
                self.assertMatch(out, "unicodePwd:: %s" % unicodePwd,
                                 "getpassword unicodePwd: out[%s]" % out)
            else:
                self.assertNotIn("unicodePwd:: %s" % unicodePwd, out)
            self.assertMatch(out, "# supplementalCredentials::: REDACTED SECRET ATTRIBUTE",
                             "getpassword '# supplementalCredentials::: REDACTED SECRET ATTRIBUTE': out[%s]" % out)
            self.assertMatch(out, "supplementalCredentials:: ",
                             "getpassword supplementalCredentials: out[%s]" % out)
            if "virtualSambaGPG:: " in out:
                self.assertMatch(out, "virtualClearTextUTF8:: %s" % virtualClearTextUTF8,
                                 "getpassword virtualClearTextUTF8: out[%s]" % out)
                self.assertMatch(out, "virtualClearTextUTF16:: %s" % virtualClearTextUTF16,
                                 "getpassword virtualClearTextUTF16: out[%s]" % out)
                self.assertMatch(out, "virtualSSHA: ",
                                 "getpassword virtualSSHA: out[%s]" % out)

            (result, out, err) = self.runsubcmd("user", "getpassword",
                                                user["name"],
                                                "--attributes=%s" % attributes,
                                                "--decrypt-samba-gpg")
            self.assertCmdSuccess(result, out, err, "Ensure getpassword runs")
            self.assertEqual(err, "Any available password returned OK\n", "getpassword without url")
            self.assertMatch(out, "sAMAccountName: %s" % (user["name"]),
                             "getpassword: 'sAMAccountName': %s out[%s]" % (user["name"], out))
            if expect_nt_hash or "virtualSambaGPG:: " in out:
                self.assertMatch(out, "unicodePwd:: %s" % unicodePwd,
                                 "getpassword unicodePwd: out[%s]" % out)
            else:
                self.assertNotIn("unicodePwd:: %s" % unicodePwd, out)
            self.assertMatch(out, "supplementalCredentials:: ",
                             "getpassword supplementalCredentials: out[%s]" % out)
            self._verify_supplementalCredentials(out)
            if "virtualSambaGPG:: " in out:
                self.assertMatch(out, "virtualClearTextUTF8:: %s" % virtualClearTextUTF8,
                                 "getpassword virtualClearTextUTF8: out[%s]" % out)
                self.assertMatch(out, "virtualClearTextUTF16:: %s" % virtualClearTextUTF16,
                                 "getpassword virtualClearTextUTF16: out[%s]" % out)
                self.assertMatch(out, "virtualSSHA: ",
                                 "getpassword virtualSSHA: out[%s]" % out)

        for user in self.users:
            newpasswd = self.random_password(16)
            (result, out, err) = self.runsubcmd("user", "setpassword",
                                                user["name"],
                                                "--newpassword=%s" % newpasswd,
                                                "--must-change-at-next-login",
                                                "-H", "ldap://%s" % os.environ["DC_SERVER"],
                                                "-U%s%%%s" % (os.environ["DC_USERNAME"], os.environ["DC_PASSWORD"]))
            self.assertCmdSuccess(result, out, err, "Ensure setpassword runs")
            self.assertEqual(err, "", "setpassword with forced change")
            self.assertMatch(out, "Changed password OK", "setpassword with forced change")

    def test_setexpiry(self):
        for user in self.users:
            twodays = time.time() + (2 * 24 * 60 * 60)

            (result, out, err) = self.runsubcmd("user", "setexpiry", user["name"],
                                                "--days=2",
                                                "-H", "ldap://%s" % os.environ["DC_SERVER"],
                                                "-U%s%%%s" % (os.environ["DC_USERNAME"], os.environ["DC_PASSWORD"]))
            self.assertCmdSuccess(result, out, err, "Can we run setexpiry with names")
            self.assertIn("Expiry for user '%s' set to 2 days." % user["name"], out)

            found = self._find_user(user["name"])

            expires = nttime2unix(int("%s" % found.get("accountExpires")))
            self.assertWithin(expires, twodays, 5, "Ensure account expires is within 5 seconds of the expected time")

        # TODO: re-enable this after the filter case is sorted out
        if "filters are broken, bail now":
            return

        # now run the expiration based on a filter
        fourdays = time.time() + (4 * 24 * 60 * 60)
        (result, out, err) = self.runsubcmd("user", "setexpiry",
                                            "--filter", "(&(objectClass=user)(company=comp2))",
                                            "--days=4",
                                            "-H", "ldap://%s" % os.environ["DC_SERVER"],
                                            "-U%s%%%s" % (os.environ["DC_USERNAME"], os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(result, out, err, "Can we run setexpiry with a filter")

        for user in self.users:
            found = self._find_user(user["name"])
            if ("%s" % found.get("company")) == "comp2":
                expires = nttime2unix(int("%s" % found.get("accountExpires")))
                self.assertWithin(expires, fourdays, 5, "Ensure account expires is within 5 seconds of the expected time")
            else:
                expires = nttime2unix(int("%s" % found.get("accountExpires")))
                self.assertWithin(expires, twodays, 5, "Ensure account expires is within 5 seconds of the expected time")

    def test_list(self):
        (result, out, err) = self.runsubcmd("user", "list",
                                            "-H", "ldap://%s" % os.environ["DC_SERVER"],
                                            "-U%s%%%s" % (os.environ["DC_USERNAME"],
                                                          os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(result, out, err, "Error running list")

        search_filter = ("(&(objectClass=user)(userAccountControl:%s:=%u))" %
                         (ldb.OID_COMPARATOR_AND, dsdb.UF_NORMAL_ACCOUNT))

        userlist = self.samdb.search(base=self.samdb.domain_dn(),
                                     scope=ldb.SCOPE_SUBTREE,
                                     expression=search_filter,
                                     attrs=["samaccountname"])

        self.assertTrue(len(userlist) > 0, "no users found in samdb")

        for userobj in userlist:
            name = str(userobj.get("samaccountname", idx=0))
            self.assertMatch(out, name,
                             "user '%s' not found" % name)

    # Test: samba-tool user list --locked-only
    # This test does not verify that the command lists the locked user, it just
    # tests that it does not list unlocked users. The funcional test, which
    # lists locked users, is located in the 'samba4.ldap.password_lockout' test
    # in source8/dsdb/tests/python/password_lockout.py
    def test_list_locked(self):
        (result, out, err) = self.runsubcmd("user", "list",
                                            "-H", "ldap://%s" % os.environ["DC_SERVER"],
                                            "-U%s%%%s" % (os.environ["DC_USERNAME"],
                                                          os.environ["DC_PASSWORD"]),
                                            "--locked-only")
        self.assertCmdSuccess(result, out, err, "Error running list")

        search_filter = ("(&(objectClass=user)(userAccountControl:%s:=%u))" %
                         (ldb.OID_COMPARATOR_AND, dsdb.UF_NORMAL_ACCOUNT))

        userlist = self.samdb.search(base=self.samdb.domain_dn(),
                                     scope=ldb.SCOPE_SUBTREE,
                                     expression=search_filter,
                                     attrs=["samaccountname"])

        for userobj in userlist:
            name = str(userobj.get("samaccountname", idx=0))
            self.assertNotIn(name, out,
                             "user '%s' is incorrectly listed as locked" % name)

    def test_list_base_dn(self):
        base_dn = "CN=Users"
        (result, out, err) = self.runsubcmd("user", "list", "-b", base_dn,
                                            "-H", "ldap://%s" % os.environ["DC_SERVER"],
                                            "-U%s%%%s" % (os.environ["DC_USERNAME"],
                                                          os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(result, out, err, "Error running list")

        search_filter = ("(&(objectClass=user)(userAccountControl:%s:=%u))" %
                         (ldb.OID_COMPARATOR_AND, dsdb.UF_NORMAL_ACCOUNT))

        userlist = self.samdb.search(base=self.samdb.normalize_dn_in_domain(base_dn),
                                     scope=ldb.SCOPE_SUBTREE,
                                     expression=search_filter,
                                     attrs=["samaccountname"])

        self.assertTrue(len(userlist) > 0, "no users found in samdb")

        for userobj in userlist:
            name = str(userobj.get("samaccountname", idx=0))
            self.assertMatch(out, name,
                             "user '%s' not found" % name)

    def test_list_full_dn(self):
        (result, out, err) = self.runsubcmd("user", "list", "--full-dn",
                                            "-H", "ldap://%s" % os.environ["DC_SERVER"],
                                            "-U%s%%%s" % (os.environ["DC_USERNAME"],
                                                          os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(result, out, err, "Error running list")

        search_filter = ("(&(objectClass=user)(userAccountControl:%s:=%u))" %
                         (ldb.OID_COMPARATOR_AND, dsdb.UF_NORMAL_ACCOUNT))

        userlist = self.samdb.search(base=self.samdb.domain_dn(),
                                     scope=ldb.SCOPE_SUBTREE,
                                     expression=search_filter,
                                     attrs=["dn"])

        self.assertTrue(len(userlist) > 0, "no users found in samdb")

        for userobj in userlist:
            name = str(userobj.get("dn", idx=0))
            self.assertMatch(out, name,
                             "user '%s' not found" % name)

    def test_list_hide_expired(self):
        expire_username = "expireUser"
        expire_user = self._randomUser({"name": expire_username})
        self._create_user(expire_user)

        (result, out, err) = self.runsubcmd(
            "user",
            "list",
            "--hide-expired",
            "-H",
            "ldap://%s" % os.environ["DC_SERVER"],
            "-U%s%%%s" % (os.environ["DC_USERNAME"],
                          os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(result, out, err, "Error running list")
        self.assertTrue(expire_username in out,
                        "user '%s' not found" % expire_username)

        # user will be expired one second ago
        self.samdb.setexpiry(
            "(sAMAccountname=%s)" % expire_username,
            -1,
            False)

        (result, out, err) = self.runsubcmd(
            "user",
            "list",
            "--hide-expired",
            "-H",
            "ldap://%s" % os.environ["DC_SERVER"],
            "-U%s%%%s" % (os.environ["DC_USERNAME"],
                          os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(result, out, err, "Error running list")
        self.assertFalse(expire_username in out,
                         "user '%s' found" % expire_username)

        self.samdb.deleteuser(expire_username)

    def test_list_hide_disabled(self):
        disable_username = "disableUser"
        disable_user = self._randomUser({"name": disable_username})
        self._create_user(disable_user)

        (result, out, err) = self.runsubcmd(
            "user",
            "list",
            "--hide-disabled",
            "-H",
            "ldap://%s" % os.environ["DC_SERVER"],
            "-U%s%%%s" % (os.environ["DC_USERNAME"],
                          os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(result, out, err, "Error running list")
        self.assertTrue(disable_username in out,
                        "user '%s' not found" % disable_username)

        self.samdb.disable_account("(sAMAccountname=%s)" % disable_username)

        (result, out, err) = self.runsubcmd(
            "user",
            "list",
            "--hide-disabled",
            "-H",
            "ldap://%s" % os.environ["DC_SERVER"],
            "-U%s%%%s" % (os.environ["DC_USERNAME"],
                          os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(result, out, err, "Error running list")
        self.assertFalse(disable_username in out,
                         "user '%s' found" % disable_username)

        self.samdb.deleteuser(disable_username)

    def test_show(self):
        for user in self.users:
            (result, out, err) = self.runsubcmd(
                "user", "show", user["name"],
                "--attributes=sAMAccountName,company",
                "-H", "ldap://%s" % os.environ["DC_SERVER"],
                "-U%s%%%s" % (os.environ["DC_USERNAME"],
                              os.environ["DC_PASSWORD"]))
            self.assertCmdSuccess(result, out, err, "Error running show")

            expected_out = """dn: CN=%s %s,CN=Users,%s
company: %s
sAMAccountName: %s

""" % (user["given-name"], user["surname"], self.samdb.domain_dn(),
                user["company"], user["name"])

            self.assertEqual(out, expected_out,
                             "Unexpected show output for user '%s'" %
                             user["name"])

            time_attrs = [
                "name", # test that invalid values are just ignored
                "whenCreated",
                "whenChanged",
                "accountExpires",
                "badPasswordTime",
                "lastLogoff",
                "lastLogon",
                "lastLogonTimestamp",
                "lockoutTime",
                "msDS-UserPasswordExpiryTimeComputed",
                "pwdLastSet",
                ]

            attrs = []
            for ta in time_attrs:
                attrs.append(ta)
                for fm in ["GeneralizedTime", "UnixTime", "TimeSpec"]:
                    attrs.append("%s;format=%s" % (ta, fm))

            (result, out, err) = self.runsubcmd(
                "user", "show", user["name"],
                "--attributes=%s" % ",".join(attrs),
                "-H", "ldap://%s" % os.environ["DC_SERVER"],
                "-U%s%%%s" % (os.environ["DC_USERNAME"],
                              os.environ["DC_PASSWORD"]))
            self.assertCmdSuccess(result, out, err,
                                  "Error running show --attributes=%s"
                                  % ",".join(attrs))

            self.assertIn(";format=GeneralizedTime", out)
            self.assertIn(";format=UnixTime", out)
            self.assertIn(";format=TimeSpec", out)

            self.assertIn("name: ", out)
            self.assertNotIn("name;format=GeneralizedTime: ", out)
            self.assertNotIn("name;format=UnixTime: ", out)
            self.assertNotIn("name;format=TimeSpec: ", out)

            self.assertIn("whenCreated: 20", out)
            self.assertIn("whenCreated;format=GeneralizedTime: 20", out)
            self.assertIn("whenCreated;format=UnixTime: 1", out)
            self.assertIn("whenCreated;format=TimeSpec: 1", out)

            self.assertIn("whenChanged: 20", out)
            self.assertIn("whenChanged;format=GeneralizedTime: 20", out)
            self.assertIn("whenChanged;format=UnixTime: 1", out)
            self.assertIn("whenChanged;format=TimeSpec: 1", out)

            self.assertIn("accountExpires: 9223372036854775807", out)
            self.assertNotIn("accountExpires;format=GeneralizedTime: ", out)
            self.assertNotIn("accountExpires;format=UnixTime: ", out)
            self.assertNotIn("accountExpires;format=TimeSpec: ", out)

            self.assertIn("badPasswordTime: 0", out)
            self.assertNotIn("badPasswordTime;format=GeneralizedTime: ", out)
            self.assertNotIn("badPasswordTime;format=UnixTime: ", out)
            self.assertNotIn("badPasswordTime;format=TimeSpec: ", out)

            self.assertIn("lastLogoff: 0", out)
            self.assertNotIn("lastLogoff;format=GeneralizedTime: ", out)
            self.assertNotIn("lastLogoff;format=UnixTime: ", out)
            self.assertNotIn("lastLogoff;format=TimeSpec: ", out)

            self.assertIn("lastLogon: 0", out)
            self.assertNotIn("lastLogon;format=GeneralizedTime: ", out)
            self.assertNotIn("lastLogon;format=UnixTime: ", out)
            self.assertNotIn("lastLogon;format=TimeSpec: ", out)

            # If a specified attribute is not available on a user object
            # it's silently omitted.
            self.assertNotIn("lastLogonTimestamp:", out)
            self.assertNotIn("lockoutTime:", out)

            self.assertIn("msDS-UserPasswordExpiryTimeComputed: 1", out)
            self.assertIn("msDS-UserPasswordExpiryTimeComputed;format=GeneralizedTime: 20", out)
            self.assertIn("msDS-UserPasswordExpiryTimeComputed;format=UnixTime: 1", out)
            self.assertIn("msDS-UserPasswordExpiryTimeComputed;format=TimeSpec: 1", out)

            self.assertIn("pwdLastSet: 1", out)
            self.assertIn("pwdLastSet;format=GeneralizedTime: 20", out)
            self.assertIn("pwdLastSet;format=UnixTime: 1", out)
            self.assertIn("pwdLastSet;format=TimeSpec: 1", out)

            out_msgs = self.samdb.parse_ldif(out)
            out_msg = next(out_msgs)[1]

            self.assertIn("whenCreated", out_msg)
            when_created_str = str(out_msg["whenCreated"][0])
            self.assertIn("whenCreated;format=GeneralizedTime", out_msg)
            self.assertEqual(str(out_msg["whenCreated;format=GeneralizedTime"][0]), when_created_str)
            when_created_time = ldb.string_to_time(when_created_str)
            self.assertIn("whenCreated;format=UnixTime", out_msg)
            self.assertEqual(str(out_msg["whenCreated;format=UnixTime"][0]), str(when_created_time))
            self.assertIn("whenCreated;format=TimeSpec", out_msg)
            self.assertEqual(str(out_msg["whenCreated;format=TimeSpec"][0]),
                             "%d.000000000" % (when_created_time))

            self.assertIn("whenChanged", out_msg)
            when_changed_str = str(out_msg["whenChanged"][0])
            self.assertIn("whenChanged;format=GeneralizedTime", out_msg)
            self.assertEqual(str(out_msg["whenChanged;format=GeneralizedTime"][0]), when_changed_str)
            when_changed_time = ldb.string_to_time(when_changed_str)
            self.assertIn("whenChanged;format=UnixTime", out_msg)
            self.assertEqual(str(out_msg["whenChanged;format=UnixTime"][0]), str(when_changed_time))
            self.assertIn("whenChanged;format=TimeSpec", out_msg)
            self.assertEqual(str(out_msg["whenChanged;format=TimeSpec"][0]),
                             "%d.000000000" % (when_changed_time))

            self.assertIn("pwdLastSet;format=GeneralizedTime", out_msg)
            pwd_last_set_str = str(out_msg["pwdLastSet;format=GeneralizedTime"][0])
            pwd_last_set_time = ldb.string_to_time(pwd_last_set_str)
            self.assertIn("pwdLastSet;format=UnixTime", out_msg)
            self.assertEqual(str(out_msg["pwdLastSet;format=UnixTime"][0]), str(pwd_last_set_time))
            self.assertIn("pwdLastSet;format=TimeSpec", out_msg)
            self.assertIn("%d." % pwd_last_set_time, str(out_msg["pwdLastSet;format=TimeSpec"][0]))
            self.assertNotIn(".000000000", str(out_msg["pwdLastSet;format=TimeSpec"][0]))

            # assert that the pwd has been set in the minute after user creation
            self.assertGreaterEqual(pwd_last_set_time, when_created_time)
            self.assertLess(pwd_last_set_time, when_created_time + 60)

            self.assertIn("msDS-UserPasswordExpiryTimeComputed;format=GeneralizedTime", out_msg)
            pwd_expires_str = str(out_msg["msDS-UserPasswordExpiryTimeComputed;format=GeneralizedTime"][0])
            pwd_expires_time = ldb.string_to_time(pwd_expires_str)
            self.assertIn("msDS-UserPasswordExpiryTimeComputed;format=UnixTime", out_msg)
            self.assertEqual(str(out_msg["msDS-UserPasswordExpiryTimeComputed;format=UnixTime"][0]), str(pwd_expires_time))
            self.assertIn("msDS-UserPasswordExpiryTimeComputed;format=TimeSpec", out_msg)
            self.assertIn("%d." % pwd_expires_time, str(out_msg["msDS-UserPasswordExpiryTimeComputed;format=TimeSpec"][0]))
            self.assertNotIn(".000000000", str(out_msg["msDS-UserPasswordExpiryTimeComputed;format=TimeSpec"][0]))

            # assert that the pwd expires after it was set
            self.assertGreater(pwd_expires_time, pwd_last_set_time)

    def test_move(self):
        full_ou_dn = str(self.samdb.normalize_dn_in_domain("OU=movetest_usr"))
        self.addCleanup(self.samdb.delete, full_ou_dn, ["tree_delete:1"])

        (result, out, err) = self.runsubcmd("ou", "add", full_ou_dn)
        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "There shouldn't be any error message")
        self.assertIn('Added ou "%s"' % full_ou_dn, out)

        for user in self.users:
            (result, out, err) = self.runsubcmd(
                "user", "move", user["name"], full_ou_dn)
            self.assertCmdSuccess(result, out, err, "Error running move")
            self.assertIn('Moved user "%s" into "%s"' %
                          (user["name"], full_ou_dn), out)

        # Should fail as users objects are in OU
        (result, out, err) = self.runsubcmd("ou", "delete", full_ou_dn)
        self.assertCmdFail(result)
        self.assertIn(("subtree_delete: Unable to delete a non-leaf node "
                       "(it has %d children)!") % len(self.users), err)

        for user in self.users:
            new_dn = "CN=Users,%s" % self.samdb.domain_dn()
            (result, out, err) = self.runsubcmd(
                "user", "move", user["name"], new_dn)
            self.assertCmdSuccess(result, out, err, "Error running move")
            self.assertIn('Moved user "%s" into "%s"' %
                          (user["name"], new_dn), out)

    def test_rename_surname_initials_givenname(self):
        """rename the existing surname and given name and add missing
        initials, then remove them, for all users"""
        for user in self.users:
            new_givenname = "new_given_name_of_" + user["name"]
            new_initials = "A"
            new_surname = "new_surname_of_" + user["name"]
            found = self._find_user(user["name"])
            old_cn = str(found.get("cn"))

            # rename given name, initials and surname
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--surname=%s" % new_surname,
                                                "--initials=%s" % new_initials,
                                                "--given-name=%s" % new_givenname)
            self.assertCmdSuccess(result, out, err)
            self.assertEqual(err, "", "Shouldn't be any error messages")
            self.assertIn('successfully', out)

            found = self._find_user(user["name"])
            self.assertEqual("%s" % found.get("givenName"), new_givenname)
            self.assertEqual("%s" % found.get("initials"), new_initials)
            self.assertEqual("%s" % found.get("sn"), new_surname)
            self.assertEqual("%s" % found.get("name"),
                             "%s %s. %s" % (new_givenname, new_initials, new_surname))
            self.assertEqual("%s" % found.get("cn"),
                             "%s %s. %s" % (new_givenname, new_initials, new_surname))

            # remove given name, initials and surname
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--surname=",
                                                "--initials=",
                                                "--given-name=")
            self.assertCmdSuccess(result, out, err)
            self.assertEqual(err, "", "Shouldn't be any error messages")
            self.assertIn('successfully', out)

            found = self._find_user(user["name"])
            self.assertEqual(found.get("givenName"), None)
            self.assertEqual(found.get("initials"), None)
            self.assertEqual(found.get("sn"), None)
            self.assertEqual("%s" % found.get("cn"), user["name"])

            # reset changes (initials are removed)
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--surname=%(surname)s" % user,
                                                "--given-name=%(given-name)s" % user)
            self.assertCmdSuccess(result, out, err)

            if old_cn:
                (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--force-new-cn=%s" % old_cn)

    def test_rename_cn_samaccountname(self):
        """rename and try to remove the cn and the samaccount of all users"""
        for user in self.users:
            new_cn = "new_cn_of_" + user["name"]
            new_samaccountname = "new_samaccount_of_" + user["name"]
            new_surname = "new_surname_of_" + user["name"]

            # rename cn
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--samaccountname=%s"
                                                 % new_samaccountname,
                                                "--force-new-cn=%s" % new_cn)
            self.assertCmdSuccess(result, out, err)
            self.assertEqual(err, "", "Shouldn't be any error messages")
            self.assertIn('successfully', out)

            found = self._find_user(new_samaccountname)
            self.assertEqual("%s" % found.get("cn"), new_cn)
            self.assertEqual("%s" % found.get("sAMAccountName"),
                             new_samaccountname)

            # changing the surname has no effect to the cn
            (result, out, err) = self.runsubcmd("user", "rename", new_samaccountname,
                                                "--surname=%s" % new_surname)
            self.assertCmdSuccess(result, out, err)

            found = self._find_user(new_samaccountname)
            self.assertEqual("%s" % found.get("cn"), new_cn)

            # trying to remove cn (throws an error)
            (result, out, err) = self.runsubcmd("user", "rename",
                                                new_samaccountname,
                                                "--force-new-cn=")
            self.assertCmdFail(result)
            self.assertIn('Failed to rename user', err)
            self.assertIn("delete protected attribute", err)

            # trying to remove the samccountname (throws an error)
            (result, out, err) = self.runsubcmd("user", "rename",
                                                new_samaccountname,
                                                "--samaccountname=")
            self.assertCmdFail(result)
            self.assertIn('Failed to rename user', err)
            self.assertIn('delete protected attribute', err)

            # reset changes (cn must be the name)
            (result, out, err) = self.runsubcmd("user", "rename", new_samaccountname,
                                                "--samaccountname=%(name)s"
                                                  % user,
                                                "--force-new-cn=%(name)s" % user)
            self.assertCmdSuccess(result, out, err)

    def test_rename_standard_cn(self):
        """reset the cn of all users to the standard"""
        for user in self.users:
            new_cn = "new_cn_of_" + user["name"]
            new_givenname = "new_given_name_of_" + user["name"]
            new_initials = "A"
            new_surname = "new_surname_of_" + user["name"]

            # set different cn
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--force-new-cn=%s" % new_cn)
            self.assertCmdSuccess(result, out, err)

            # remove given name, initials and surname
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--surname=",
                                                "--initials=",
                                                "--given-name=")
            self.assertCmdSuccess(result, out, err)

            # reset the CN (no given name, initials or surname --> samaccountname)
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--reset-cn")

            self.assertCmdSuccess(result, out, err)
            self.assertEqual(err, "", "Shouldn't be any error messages")
            self.assertIn('successfully', out)

            found = self._find_user(user["name"])
            self.assertEqual("%s" % found.get("cn"), user["name"])

            # set given name, initials and surname and set different cn
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--force-new-cn=%s" % new_cn,
                                                "--surname=%s" % new_surname,
                                                "--initials=%s" % new_initials,
                                                "--given-name=%s" % new_givenname)
            self.assertCmdSuccess(result, out, err)

            # reset the CN (given name, initials or surname are given --> given name)
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--reset-cn")

            self.assertCmdSuccess(result, out, err)
            self.assertEqual(err, "", "Shouldn't be any error messages")
            self.assertIn('successfully', out)

            found = self._find_user(user["name"])
            self.assertEqual("%s" % found.get("cn"),
                             "%s %s. %s" % (new_givenname, new_initials, new_surname))

            # reset changes
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--reset-cn",
                                                "--initials=",
                                                "--surname=%(surname)s" % user,
                                                "--given-name=%(given-name)s" % user)
            self.assertCmdSuccess(result, out, err)

    def test_rename_mailaddress_displayname(self):
        for user in self.users:
            new_mail = "new_mailaddress_of_" + user["name"]
            new_displayname = "new displayname of " + user["name"]

            # change mail and displayname
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--mail-address=%s"
                                                  % new_mail,
                                                "--display-name=%s"
                                                  % new_displayname)
            self.assertCmdSuccess(result, out, err)
            self.assertEqual(err, "", "Shouldn't be any error messages")
            self.assertIn('successfully', out)

            found = self._find_user(user["name"])
            self.assertEqual("%s" % found.get("mail"), new_mail)
            self.assertEqual("%s" % found.get("displayName"), new_displayname)

            # remove mail and displayname
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--mail-address=",
                                                "--display-name=")
            self.assertCmdSuccess(result, out, err)
            self.assertEqual(err, "", "Shouldn't be any error messages")
            self.assertIn('successfully', out)

            found = self._find_user(user["name"])
            self.assertEqual(found.get("mail"), None)
            self.assertEqual(found.get("displayName"), None)

    def test_rename_upn(self):
        """rename upn of all users"""
        for user in self.users:
            found = self._find_user(user["name"])
            old_upn = "%s" % found.get("userPrincipalName")
            valid_suffix = old_upn.split('@')[1]   # samba.example.com

            valid_new_upn = "new_%s@%s" % (user["name"], valid_suffix)
            invalid_new_upn = "%s@invalid.suffix" + user["name"]

            # trying to set invalid upn
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--upn=%s"
                                                  % invalid_new_upn)
            self.assertCmdFail(result)
            self.assertIn('is not a valid upn', err)

            # set valid upn
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--upn=%s"
                                                  % valid_new_upn)
            self.assertCmdSuccess(result, out, err)
            self.assertEqual(err, "", "Shouldn't be any error messages")
            self.assertIn('successfully', out)

            found = self._find_user(user["name"])
            self.assertEqual("%s" % found.get("userPrincipalName"), valid_new_upn)

            # trying to remove upn
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--upn=%s")
            self.assertCmdFail(result)
            self.assertIn('is not a valid upn', err)

            # reset upn
            (result, out, err) = self.runsubcmd("user", "rename", user["name"],
                                                "--upn=%s" % old_upn)
            self.assertCmdSuccess(result, out, err)

    def test_getpwent(self):
        try:
            import pwd
        except ImportError:
            self.skipTest("Skipping getpwent test, no 'pwd' module available")
            return

        # get the current user's data for the test
        uid = os.geteuid()
        try:
            u = pwd.getpwuid(uid)
        except KeyError:
            self.skipTest("Skipping getpwent test, current EUID not found in NSS")
            return


# samba-tool user create command didn't support users with empty gecos if none is
# specified on the command line and the user hasn't one in the passwd file it
# will fail, so let's add some contents

        gecos = u[4]
        if (gecos is None or len(gecos) == 0):
            gecos = "Foo GECOS"
        user = self._randomPosixUser({
                        "name": u[0],
                        "uid": u[0],
                        "uidNumber": u[2],
                        "gidNumber": u[3],
                        "gecos": gecos,
                        "loginShell": u[6],
                        })

        # Remove user if it already exists
        if self._find_user(u[0]):
            self.runsubcmd("user", "delete", u[0])
        # check if --rfc2307-from-nss sets the same values as we got from pwd.getpwuid()
        (result, out, err) = self.runsubcmd("user", "create", user["name"], user["password"],
                                            "--surname=%s" % user["surname"],
                                            "--given-name=%s" % user["given-name"],
                                            "--job-title=%s" % user["job-title"],
                                            "--department=%s" % user["department"],
                                            "--description=%s" % user["description"],
                                            "--company=%s" % user["company"],
                                            "--gecos=%s" % user["gecos"],
                                            "--rfc2307-from-nss",
                                            "-H", "ldap://%s" % os.environ["DC_SERVER"],
                                            "-U%s%%%s" % (os.environ["DC_USERNAME"], os.environ["DC_PASSWORD"]))

        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.assertIn("User '%s' added successfully" % user["name"], out)

        self._check_posix_user(user)
        self.runsubcmd("user", "delete", user["name"])

        # Check if overriding the attributes from NSS with explicit values works
        #
        # get a user with all random posix attributes
        user = self._randomPosixUser({"name": u[0]})

        # Remove user if it already exists
        if self._find_user(u[0]):
            self.runsubcmd("user", "delete", u[0])
        # create a user with posix attributes from nss but override all of them with the
        # random ones just obtained
        (result, out, err) = self.runsubcmd("user", "create", user["name"], user["password"],
                                            "--surname=%s" % user["surname"],
                                            "--given-name=%s" % user["given-name"],
                                            "--job-title=%s" % user["job-title"],
                                            "--department=%s" % user["department"],
                                            "--description=%s" % user["description"],
                                            "--company=%s" % user["company"],
                                            "--rfc2307-from-nss",
                                            "--gecos=%s" % user["gecos"],
                                            "--login-shell=%s" % user["loginShell"],
                                            "--uid=%s" % user["uid"],
                                            "--uid-number=%s" % user["uidNumber"],
                                            "--gid-number=%s" % user["gidNumber"],
                                            "-H", "ldap://%s" % os.environ["DC_SERVER"],
                                            "-U%s%%%s" % (os.environ["DC_USERNAME"], os.environ["DC_PASSWORD"]))

        self.assertCmdSuccess(result, out, err)
        self.assertEqual(err, "", "Shouldn't be any error messages")
        self.assertIn("User '%s' added successfully" % user["name"], out)

        self._check_posix_user(user)
        self.runsubcmd("user", "delete", user["name"])

    # Test: samba-tool user unlock
    # This test does not verify that the command unlocks the user, it just
    # tests the command itself. The unlock test, which unlocks locked users,
    # is located in the 'samba4.ldap.password_lockout' test in
    # source4/dsdb/tests/python/password_lockout.py
    def test_unlock(self):

        # try to unlock a nonexistent user, this should fail
        nonexistentusername = "userdoesnotexist"
        (result, out, err) = self.runsubcmd(
            "user", "unlock", nonexistentusername)
        self.assertCmdFail(result, "Ensure that unlock nonexistent user fails")
        self.assertIn("Failed to unlock user '%s'" % nonexistentusername, err)
        self.assertIn("Unable to find user", err)

        # try to unlock with insufficient permissions, this should fail
        unprivileged_username = "unprivilegedunlockuser"
        unlocktest_username = "usertounlock"

        self.runsubcmd("user", "add", unprivileged_username, "Passw0rd")
        self.runsubcmd("user", "add", unlocktest_username, "Passw0rd")

        (result, out, err) = self.runsubcmd(
            "user", "unlock", unlocktest_username,
            "-H", "ldap://%s" % os.environ["DC_SERVER"],
            "-U%s%%%s" % (unprivileged_username,
                          "Passw0rd"))
        self.assertCmdFail(result, "Fail with LDAP_INSUFFICIENT_ACCESS_RIGHTS")
        self.assertIn("Failed to unlock user '%s'" % unlocktest_username, err)
        self.assertIn("LDAP error 50 LDAP_INSUFFICIENT_ACCESS_RIGHTS", err)

        self.runsubcmd("user", "delete", unprivileged_username)
        self.runsubcmd("user", "delete", unlocktest_username)

        # run unlock against test users
        for user in self.users:
            (result, out, err) = self.runsubcmd(
                "user", "unlock", user["name"])
            self.assertCmdSuccess(result, out, err, "Error running user unlock")
            self.assertEqual(err, "", "Shouldn't be any error messages")

    def test_disable_remove_supplemental_groups(self):
        """disable user and remove supplemental groups"""
        username = "userRemoveGroups"
        user = self._randomUser({"name": username})
        self._create_user(user)

        usergroups = self._get_groups(username)
        self.assertTrue(len(usergroups) == 1, "exactly one membership expected")
        self.assertEqual(usergroups[0],
                         "Domain Users",
                         "Unexpected groupmembership")

        self._add_groupmember("Domain Admins", username)
        self._add_groupmember("Print Operators", username)

        usergroups = self._get_groups(username)
        self.assertTrue(len(usergroups) == 3, "exactly 3 memberships expected")

        (result, out, err) = self.runsubcmd(
            "user", "disable", username,
            "--remove-supplemental-groups",
            "-H", "ldap://%s" % os.environ["DC_SERVER"],
            "-U%s%%%s" % (os.environ["DC_USERNAME"],
            os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(
            result, out, err,
            "Error running user disable --remove-supplemental-groups")
        self.assertEqual(err, "",
                         "Shouldn't be any error messages from user disable")

        usergroups = self._get_groups(username)
        self.assertTrue(len(usergroups) == 1, "exactly one membership expected")
        self.assertEqual(usergroups[0], "Domain Users",
                         "Unexpected groupmembership")

    def _randomUser(self, base=None):
        """create a user with random attribute values, you can specify base attributes"""
        if base is None:
            base = {}
        user = {
            "name": self.randomName(),
            "password": self.random_password(16),
            "surname": self.randomName(),
            "given-name": self.randomName(),
            "job-title": self.randomName(),
            "department": self.randomName(),
            "company": self.randomName(),
            "description": self.randomName(count=100),
            "createUserFn": self._create_user,
            "checkUserFn": self._check_user,
        }
        user.update(base)
        return user

    def _randomPosixUser(self, base=None):
        """create a user with random attribute values and additional RFC2307
        attributes, you can specify base attributes"""
        if base is None:
            base = {}
        user = self._randomUser({})
        user.update(base)
        posixAttributes = {
            "uid": self.randomName(),
            "loginShell": self.randomName(),
            "gecos": self.randomName(),
            "uidNumber": self.randomXid(),
            "gidNumber": self.randomXid(),
            "createUserFn": self._create_posix_user,
            "checkUserFn": self._check_posix_user,
        }
        user.update(posixAttributes)
        user.update(base)
        return user

    def _randomUnixUser(self, base=None):
        """create a user with random attribute values and additional RFC2307
        attributes, you can specify base attributes"""
        if base is None:
            base = {}
        user = self._randomUser({})
        user.update(base)
        posixAttributes = {
            "uidNumber": self.randomXid(),
            "gidNumber": self.randomXid(),
            "uid": self.randomName(),
            "loginShell": self.randomName(),
            "gecos": self.randomName(),
            "createUserFn": self._create_unix_user,
            "checkUserFn": self._check_unix_user,
        }
        user.update(posixAttributes)
        user.update(base)
        return user

    def _check_user(self, user):
        """ check if a user from SamDB has the same attributes as its template """
        found = self._find_user(user["name"])

        self.assertEqual("%s" % found.get("name"), "%(given-name)s %(surname)s" % user)
        self.assertEqual("%s" % found.get("title"), user["job-title"])
        self.assertEqual("%s" % found.get("company"), user["company"])
        self.assertEqual("%s" % found.get("description"), user["description"])
        self.assertEqual("%s" % found.get("department"), user["department"])

    def _check_posix_user(self, user):
        """ check if a posix_user from SamDB has the same attributes as its template """
        found = self._find_user(user["name"])

        self.assertEqual("%s" % found.get("loginShell"), user["loginShell"])
        self.assertEqual("%s" % found.get("gecos"), user["gecos"])
        self.assertEqual("%s" % found.get("uidNumber"), "%s" % user["uidNumber"])
        self.assertEqual("%s" % found.get("gidNumber"), "%s" % user["gidNumber"])
        self.assertEqual("%s" % found.get("uid"), user["uid"])
        self._check_user(user)

    def _check_unix_user(self, user):
        """ check if a unix_user from SamDB has the same attributes as its
template """
        found = self._find_user(user["name"])

        self.assertEqual("%s" % found.get("loginShell"), user["loginShell"])
        self.assertEqual("%s" % found.get("gecos"), user["gecos"])
        self.assertEqual("%s" % found.get("uidNumber"), "%s" %
                          user["uidNumber"])
        self.assertEqual("%s" % found.get("gidNumber"), "%s" %
                          user["gidNumber"])
        self.assertEqual("%s" % found.get("uid"), user["uid"])
        self.assertIn('/home/test/', "%s" % found.get("unixHomeDirectory"))
        self._check_user(user)

    def _create_user(self, user):
        return self.runsubcmd("user", "add", user["name"], user["password"],
                              "--surname=%s" % user["surname"],
                              "--given-name=%s" % user["given-name"],
                              "--job-title=%s" % user["job-title"],
                              "--department=%s" % user["department"],
                              "--description=%s" % user["description"],
                              "--company=%s" % user["company"],
                              "-H", "ldap://%s" % os.environ["DC_SERVER"],
                              "-U%s%%%s" % (os.environ["DC_USERNAME"], os.environ["DC_PASSWORD"]))

    def _create_posix_user(self, user):
        """ create a new user with RFC2307 attributes """
        return self.runsubcmd("user", "create", user["name"], user["password"],
                              "--surname=%s" % user["surname"],
                              "--given-name=%s" % user["given-name"],
                              "--job-title=%s" % user["job-title"],
                              "--department=%s" % user["department"],
                              "--description=%s" % user["description"],
                              "--company=%s" % user["company"],
                              "--gecos=%s" % user["gecos"],
                              "--login-shell=%s" % user["loginShell"],
                              "--uid=%s" % user["uid"],
                              "--uid-number=%s" % user["uidNumber"],
                              "--gid-number=%s" % user["gidNumber"],
                              "-H", "ldap://%s" % os.environ["DC_SERVER"],
                              "-U%s%%%s" % (os.environ["DC_USERNAME"], os.environ["DC_PASSWORD"]))

    def _create_unix_user(self, user):
        """ Add RFC2307 attributes to a user"""
        self._create_user(user)
        return self.runsubcmd("user", "addunixattrs", user["name"],
                              "%s" % user["uidNumber"],
                              "--gid-number=%s" % user["gidNumber"],
                              "--gecos=%s" % user["gecos"],
                              "--login-shell=%s" % user["loginShell"],
                              "--uid=%s" % user["uid"],
                              "-H", "ldap://%s" % os.environ["DC_SERVER"],
                              "-U%s%%%s" % (os.environ["DC_USERNAME"],
                                            os.environ["DC_PASSWORD"]))

    def _find_user(self, name):
        search_filter = "(&(sAMAccountName=%s)(objectCategory=%s,%s))" % (ldb.binary_encode(name), "CN=Person,CN=Schema,CN=Configuration", self.samdb.domain_dn())
        userlist = self.samdb.search(base=self.samdb.domain_dn(),
                                     scope=ldb.SCOPE_SUBTREE,
                                     expression=search_filter)
        if userlist:
            return userlist[0]
        else:
            return None

    def _add_groupmember(self, group, user):
        (result, out, err) =  self.runsubcmd(
            "group", "addmembers", group, user,
            "-H", "ldap://%s" % os.environ["DC_SERVER"],
            "-U%s%%%s" % (os.environ["DC_USERNAME"],
                          os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(
            result, out, err, "Error running group addmembers")
        self.assertEqual(
            err,
            "",
            "Shouldn't be any error messages from group addmembers")

        return out.rstrip().split("\n")

    def _remove_groupmember(self, group, user):
        (result, out, err) = self.runsubcmd(
            "group", "removemembers", group, user,
            "-H", "ldap://%s" % os.environ["DC_SERVER"],
            "-U%s%%%s" % (os.environ["DC_USERNAME"],
                          os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(
            result, out, err, "Error running group removemembers")
        self.assertEqual(
            err,
            "",
            "Shouldn't be any error messages from group removemembers")

        return out.rstrip().split("\n")

    def _get_groups(self, user):
        (result, out, err) = self.runsubcmd(
            "user", "getgroups", user,
            "-H", "ldap://%s" % os.environ["DC_SERVER"],
            "-U%s%%%s" % (os.environ["DC_USERNAME"],
            os.environ["DC_PASSWORD"]))
        self.assertCmdSuccess(result, out, err, "Error running user getgroups")
        self.assertEqual(err,
                         "",
                         "Shouldn't be any error messages from user getgroups")

        return out.rstrip().split("\n")



#!/usr/local/bin/python3.14
# Unix SMB/CIFS implementation.
# Copyright (C) Stefan Metzmacher 2020
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os

sys.path.insert(0, 'bin/python')
os.environ['PYTHONUNBUFFERED'] = '1'

from samba.dcerpc import krb5pac
from samba.ndr import ndr_unpack
from samba.tests import DynamicTestCase
from samba.tests.krb5.kdc_base_test import KDCBaseTest

global_asn1_print = False
global_hexdump = False


@DynamicTestCase
class PacAlignTests(KDCBaseTest):

    base_name = 'krbpac'

    @classmethod
    def setUpDynamicTestCases(cls):
        for length in range(len(cls.base_name), 21):
            cls.generate_dynamic_test('test_pac_align',
                                      f'{length}_chars',
                                      length)

    def setUp(self):
        super().setUp()
        self.do_asn1_print = global_asn1_print
        self.do_hexdump = global_hexdump

    def _test_pac_align_with_args(self, length):
        samdb = self.get_samdb()

        account_name = self.base_name + 'a' * (length - len(self.base_name))
        creds, _ = self.create_account(samdb, account_name)

        tgt = self.get_tgt(creds, expect_pac=True)

        pac_data = self.get_ticket_pac(tgt)
        self.assertIsNotNone(pac_data)

        self.assertEqual(0, len(pac_data) & 7)

        pac = ndr_unpack(krb5pac.PAC_DATA_RAW, pac_data)
        for pac_buffer in pac.buffers:
            buffer_type = pac_buffer.type
            buffer_size = pac_buffer.ndr_size

            with self.subTest(buffer_type=buffer_type):
                if buffer_type == krb5pac.PAC_TYPE_LOGON_NAME:
                    self.assertEqual(length * 2 + 10, buffer_size)
                elif buffer_type == krb5pac.PAC_TYPE_REQUESTER_SID:
                    self.assertEqual(28, buffer_size)
                elif buffer_type in {krb5pac.PAC_TYPE_SRV_CHECKSUM,
                                     krb5pac.PAC_TYPE_KDC_CHECKSUM,
                                     krb5pac.PAC_TYPE_TICKET_CHECKSUM}:
                    self.assertEqual(0, buffer_size & 3,
                                     f'buffer type was: {buffer_type}, '
                                     f'buffer size was: {buffer_size}')
                else:
                    self.assertEqual(0, buffer_size & 7,
                                     f'buffer type was: {buffer_type}, '
                                     f'buffer size was: {buffer_size}')

                rounded_len = (buffer_size + 7) & ~7
                self.assertEqual(rounded_len, len(pac_buffer.info.remaining))


if __name__ == '__main__':
    global_asn1_print = False
    global_hexdump = False
    import unittest
    unittest.main()



#
# samba-gpupdate enhanced logging
#
# Copyright (C) 2019-2020 BaseALT Ltd.
# Copyright (C) David Mulder <dmulder@samba.org> 2022
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import logging
import gettext
import random
import sys

logger = logging.getLogger("gp")


def logger_init(name, log_level):
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.setLevel(logging.CRITICAL)
    if log_level == 1:
        logger.setLevel(logging.ERROR)
    elif log_level == 2:
        logger.setLevel(logging.WARNING)
    elif log_level == 3:
        logger.setLevel(logging.INFO)
    elif log_level >= 4:
        logger.setLevel(logging.DEBUG)

class slogm(object):
    """
    Structured log message class
    """
    def __init__(self, message, kwargs=None):
        if kwargs is None:
            kwargs = {}
        self.message = message
        self.kwargs = kwargs
        if not isinstance(self.kwargs, dict):
            self.kwargs = { 'val': self.kwargs }

    def __str__(self):
        now = str(datetime.datetime.now().isoformat(sep=' ', timespec='milliseconds'))
        args = dict()
        args.update(self.kwargs)
        result = '{}|{} | {}'.format(now, self.message, args)

        return result

def message_with_code(mtype, message):
    random.seed(message)
    code = random.randint(0, 99999)
    return '[' + mtype + str(code).rjust(5, '0') + ']| ' + \
           gettext.gettext(message)

class log(object):
    @staticmethod
    def info(message, data=None):
        if data is None:
            data = {}
        msg = message_with_code('I', message)
        logger.info(slogm(msg, data))
        return msg

    @staticmethod
    def warning(message, data=None):
        if data is None:
            data = {}
        msg = message_with_code('W', message)
        logger.warning(slogm(msg, data))
        return msg

    @staticmethod
    def warn(message, data=None):
        if data is None:
            data = {}
        return log.warning(message, data)

    @staticmethod
    def error(message, data=None):
        if data is None:
            data = {}
        msg = message_with_code('E', message)
        logger.error(slogm(msg, data))
        return msg

    @staticmethod
    def fatal(message, data=None):
        if data is None:
            data = {}
        msg = message_with_code('F', message)
        logger.fatal(slogm(msg, data))
        return msg

    @staticmethod
    def debug(message, data=None):
        if data is None:
            data = {}
        msg = message_with_code('D', message)
        logger.debug(slogm(msg, data))
        return msg



# user management
#
# set user expiry
#
# Copyright Jelmer Vernooij 2010 <jelmer@samba.org>
# Copyright Theresa Halloran 2011 <theresahalloran@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import samba.getopt as options
from samba import ldb
from samba.auth import system_session
from samba.netcmd import Command, CommandError, Option
from samba.samdb import SamDB


class cmd_user_setexpiry(Command):
    """Set the expiration of a user account.

The user can either be specified by their sAMAccountName or using the --filter option.

When a user account expires, it becomes disabled and the user is unable to logon.  The administrator may issue the samba-tool user enable command to enable the account for logon.  The permissions and memberships associated with the account are retained when the account is enabled.

The command may be run from the root userid or another authorized userid.  The -H or --URL= option can be used to execute the command on a remote server.

Example1:
samba-tool user setexpiry User1 --days=20 --URL=ldap://samba.samdom.example.com --username=administrator --password=passw1rd

Example1 shows how to set the expiration of an account in a remote LDAP server.  The --URL parameter is used to specify the remote target server.  The --username= and --password= options are used to pass the username and password of a user that exists on the remote server and is authorized to update that server.

Example2:
sudo samba-tool user setexpiry User2 --noexpiry

Example2 shows how to set the account expiration of user User2 so it will never expire.  The user in this example resides on the  local server.   sudo is used so a user may run the command as root.

Example3:
samba-tool user setexpiry --days=20 --filter='(samaccountname=User3)'

Example3 shows how to set the account expiration date to end of day 20 days from the current day.  The username or sAMAccountName is specified using the --filter= parameter and the username in this example is User3.

Example4:
samba-tool user setexpiry --noexpiry User4
Example4 shows how to set the account expiration so that it will never expire.  The username and sAMAccountName in this example is User4.

"""
    synopsis = "%prog (<username>|--filter <filter>) [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "versionopts": options.VersionOptions,
        "credopts": options.CredentialsOptions,
    }

    takes_options = [
        Option("-H", "--URL", help="LDB URL for database or target server", type=str,
               metavar="URL", dest="H"),
        Option("--filter", help="LDAP Filter to set password on", type=str),
        Option("--days", help="Days to expiry", type=int, default=0),
        Option("--noexpiry", help="Password does never expire", action="store_true", default=False),
    ]

    takes_args = ["username?"]

    def run(self, username=None, sambaopts=None, credopts=None,
            versionopts=None, H=None, filter=None, days=None, noexpiry=None):
        if username is None and filter is None:
            raise CommandError("Either the username or '--filter' must be specified!")

        if filter is None:
            filter = "(&(objectClass=user)(sAMAccountName=%s))" % (ldb.binary_encode(username))

        lp = sambaopts.get_loadparm()
        creds = credopts.get_credentials(lp)

        samdb = SamDB(url=H, session_info=system_session(),
                      credentials=creds, lp=lp)

        try:
            samdb.setexpiry(filter, days * 24 * 3600, no_expiry_req=noexpiry)
        except Exception as msg:
            # FIXME: Catch more specific exception
            raise CommandError("Failed to set expiry for user '%s': %s" % (
                username or filter, msg))
        if noexpiry:
            self.outf.write("Expiry for user '%s' disabled.\n" % (
                username or filter))
        else:
            self.outf.write("Expiry for user '%s' set to %u days.\n" % (
                username or filter, days))



#!/usr/local/bin/python3.14
# Unix SMB/CIFS implementation.
# Copyright (C) Catalyst.Net Ltd 2023
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os

sys.path.insert(0, 'bin/python')
os.environ['PYTHONUNBUFFERED'] = '1'

from samba.dcerpc import gmsa
from samba.ndr import ndr_pack, ndr_unpack
import samba.tests


class GmsaTests(samba.tests.TestCase):
    managed_password_blob = (
        b'\x01\x00\x00\x00"\x01\x00\x00\x10\x00\x00\x00\x12\x01\x1a\x01g\x86W\xa1'
        b'\x13nT\x7fF\xeey\x88\xc8\x08\xd9\x04\xed\x0eK\x05\x92\xf8\x9e\xb8+\xd2\x92h'
        b'Xg\xc3\x11\x9d\xd6\xea\xae\xf5\x81\n\x1a\xa4\xe0\x8eI|\xc3\x11c'
        b'\xb2\xe7\x99\xe6\xeaf\xe3\x02,\x10\x0b\xf5\x95\x85\xa3FBt\xeb\xad$\x88\xfc('
        b'\xac\xbd\x10\xa9\xb4M\xdeCjm5\xff\xf0\xe9Z\xe7\x906\t\xe8%"\n\xd3\r\xb6\xa8k'
        b'\xb5D\xfa4\x0f\x86M-8\x95\x19=@\x07\xdfrG\x8dq\xce?x\x9b\xb19\xc4\xc1\xcf'
        b"\xfdm9\x94\x8c\n\xfaje\xe3\xf5\xf8\xf9\r\x8cp\xf7',\xe6Z?c'\x93\xeb\x0eF"
        b'\x97\xe5v\xc2\x1f6\xacU\xf4\x16z"\xb4\xeb\xb2Y<-"\xdcJ\xc8\xd4\xcaE_)\x9a'
        b'\x18+\x8dM\x8d\xd1#-\xde\x1e\xfe:\xca\xf1K\x13tS\x19_EE_]H\xa0\xc4A'
        b'\x91;\x80\xf9MF\x96\xb1q7\x9bZ\xc3\xb0,P\x1c\xf8\xe1kC\xbe\xac\xa5"cA\x1d'
        b'\\\xf7r\xe7c\xe8\xd2\x9ap\xa1)>r\x18\xa1\xe3\x00\x00t\x95\x01i\x80\x17'
        b'\x00\x00t71\xb6\x7f\x17\x00\x00'
    )

    current_password = (
        b'g\x86W\xa1\x13nT\x7fF\xeey\x88\xc8\x08\xd9\x04\xed\x0eK\x05\x92\xf8\x9e\xb8'
        b'+\xd2\x92hXg\xc3\x11\x9d\xd6\xea\xae\xf5\x81\n\x1a\xa4\xe0\x8eI|\xc3\x11c'
        b'\xb2\xe7\x99\xe6\xeaf\xe3\x02,\x10\x0b\xf5\x95\x85\xa3FBt\xeb\xad$\x88\xfc('
        b'\xac\xbd\x10\xa9\xb4M\xdeCjm5\xff\xf0\xe9Z\xe7\x906\t\xe8%"\n\xd3\r\xb6\xa8k'
        b'\xb5D\xfa4\x0f\x86M-8\x95\x19=@\x07\xdfrG\x8dq\xce?x\x9b\xb19\xc4\xc1\xcf'
        b"\xfdm9\x94\x8c\n\xfaje\xe3\xf5\xf8\xf9\r\x8cp\xf7',\xe6Z?c'\x93\xeb\x0eF"
        b'\x97\xe5v\xc2\x1f6\xacU\xf4\x16z"\xb4\xeb\xb2Y<-"\xdcJ\xc8\xd4\xcaE_)\x9a'
        b'\x18+\x8dM\x8d\xd1#-\xde\x1e\xfe:\xca\xf1K\x13tS\x19_EE_]H\xa0\xc4A'
        b'\x91;\x80\xf9MF\x96\xb1q7\x9bZ\xc3\xb0,P\x1c\xf8\xe1kC\xbe\xac\xa5"cA\x1d'
        b'\\\xf7r\xe7c\xe8\xd2\x9ap\xa1)>r\x18\xa1\xe3'
    )

    query_interval = 0x178069019574
    unchanged_interval = 0x177fb6313774

    def test_managed_password_blob_unpack(self):
        """Unpack a GMSA Managed Password blob and check its fields."""

        managed_password = ndr_unpack(gmsa.MANAGEDPASSWORD_BLOB,
                                      self.managed_password_blob)

        self.assertEqual(1, managed_password.version)
        self.assertEqual(0, managed_password.reserved)
        self.assertEqual(len(self.managed_password_blob),
                         managed_password.length)

        self.assertEqual(self.current_password,
                         managed_password.passwords.current)
        self.assertIsNone(managed_password.passwords.previous)

        self.assertEqual(self.query_interval,
                         managed_password.passwords.query_interval)
        self.assertEqual(self.unchanged_interval,
                         managed_password.passwords.unchanged_interval)

    def test_managed_password_blob_pack(self):
        """Create a GMSA Managed Password blob and test that it packs to the
        blob we expect."""

        managed_password = gmsa.MANAGEDPASSWORD_BLOB()

        managed_password.passwords.current = self.current_password
        managed_password.passwords.query_interval = self.query_interval
        managed_password.passwords.unchanged_interval = self.unchanged_interval

        self.assertEqual(self.managed_password_blob,
                         ndr_pack(managed_password))


if __name__ == '__main__':
    import unittest

    unittest.main()



# vgp_openssh_ext samba group policy
# Copyright (C) David Mulder <dmulder@suse.com> 2020
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
from io import BytesIO
from samba.gp.gpclass import gp_xml_ext, gp_file_applier
from samba.common import get_bytes

intro = b'''
### autogenerated by samba
#
# This file is generated by the vgp_openssh_ext Group Policy
# Client Side Extension. To modify the contents of this file,
# modify the appropriate Group Policy objects which apply
# to this machine. DO NOT MODIFY THIS FILE DIRECTLY.
#

'''

# For each key value pair in sshd_config, the first obtained value will be
# used. We must insert config files in reverse, so that the last applied policy
# takes precedence.
def select_next_conf(directory):
    configs = [re.match(r'(\d+)', f) for f in os.listdir(directory)]
    conf_ids = [int(m.group(1)) for m in configs if m]
    conf_ids.append(9000000000) # The starting node
    conf_id = min(conf_ids)-1
    return os.path.join(directory, '%010d_gp.conf' % conf_id)

class vgp_openssh_ext(gp_xml_ext, gp_file_applier):
    def __str__(self):
        return 'VGP/Unix Settings/OpenSSH'

    def process_group_policy(self, deleted_gpo_list, changed_gpo_list,
            cfg_dir='/etc/ssh/sshd_config.d'):
        for guid, settings in deleted_gpo_list:
            if str(self) in settings:
                for attribute, sshd_config in settings[str(self)].items():
                    self.unapply(guid, attribute, sshd_config)

        for gpo in changed_gpo_list:
            if gpo.file_sys_path:
                xml = 'MACHINE/VGP/VTLA/SshCfg/SshD/manifest.xml'
                path = os.path.join(gpo.file_sys_path, xml)
                xml_conf = self.parse(path)
                if not xml_conf:
                    continue
                policy = xml_conf.find('policysetting')
                data = policy.find('data')
                configfile = data.find('configfile')
                for configsection in configfile.findall('configsection'):
                    if configsection.find('sectionname').text:
                        continue
                    settings = {}
                    for kv in configsection.findall('keyvaluepair'):
                        settings[kv.find('key')] = kv.find('value')
                    raw = BytesIO()
                    for k, v in settings.items():
                        raw.write(b'%s %s\n' %
                                  (get_bytes(k.text), get_bytes(v.text)))
                    # Each GPO applies only one set of OpenSSH settings, in a
                    # single file, so the attribute does not need uniqueness.
                    attribute = self.generate_attribute(gpo.name)
                    # The value hash is generated from the raw data we will
                    # write to the OpenSSH settings file, ensuring any changes
                    # to this GPO will cause the file to be rewritten.
                    value_hash = self.generate_value_hash(raw.getvalue())
                    if not os.path.isdir(cfg_dir):
                        os.mkdir(cfg_dir, 0o640)
                    def applier_func(cfg_dir, raw):
                        filename = select_next_conf(cfg_dir)
                        f = open(filename, 'wb')
                        f.write(intro)
                        f.write(raw.getvalue())
                        os.chmod(filename, 0o640)
                        f.close()
                        return [filename]
                    self.apply(gpo.name, attribute, value_hash, applier_func,
                               cfg_dir, raw)
                    raw.close()

    def rsop(self, gpo):
        output = {}
        if gpo.file_sys_path:
            xml = 'MACHINE/VGP/VTLA/SshCfg/SshD/manifest.xml'
            path = os.path.join(gpo.file_sys_path, xml)
            xml_conf = self.parse(path)
            if not xml_conf:
                return output
            policy = xml_conf.find('policysetting')
            data = policy.find('data')
            configfile = data.find('configfile')
            for configsection in configfile.findall('configsection'):
                if configsection.find('sectionname').text:
                    continue
                for kv in configsection.findall('keyvaluepair'):
                    if str(self) not in output.keys():
                        output[str(self)] = {}
                    output[str(self)][kv.find('key').text] = \
                        kv.find('value').text
        return output



# -*- coding: utf-8 -*-
# Unix SMB/CIFS implementation. Tests for smb manipulation
# Copyright (C) David Mulder <dmulder@suse.com> 2018
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import samba
import os
import random
import sys
from samba import NTSTATUSError
from samba.ntstatus import (NT_STATUS_OBJECT_NAME_NOT_FOUND,
                            NT_STATUS_OBJECT_PATH_NOT_FOUND)
from samba.samba3 import libsmb_samba_internal as libsmb
from samba.samba3 import param as s3param

realm = os.environ.get('REALM')
domain_dir = realm.lower() + '/'
test_contents = 'abcd' * 256
utf_contents = u'Süßigkeiten Äpfel ' * 128
test_literal_bytes_embed_nulls = b'\xff\xfe\x14\x61\x00\x00\x62\x63\x64' * 256
binary_contents = b'\xff\xfe'
binary_contents = binary_contents + "Hello cruel world of python3".encode('utf8') * 128
test_dir = os.path.join(domain_dir, 'testing_%d' % random.randint(0, 0xFFFF))
test_file = os.path.join(test_dir, 'testing').replace('/', '\\')


class SMBTests(samba.tests.TestCase):
    def setUp(self):
        super().setUp()
        self.server = os.environ["SERVER"]
        creds = self.insta_creds(template=self.get_credentials())

        # create an SMB connection to the server
        lp = s3param.get_context()
        lp.load(os.getenv("SMB_CONF_PATH"))
        self.smb_conn = libsmb.Conn(self.server, "sysvol", lp, creds)

        self.smb_conn.mkdir(test_dir)

    def tearDown(self):
        super().tearDown()
        try:
            self.smb_conn.deltree(test_dir)
        except:
            pass

    def test_list(self):
        # check a basic listing returns the items we expect
        ls = [f['name'] for f in self.smb_conn.list(domain_dir)]
        self.assertIn('scripts', ls,
                      msg='"scripts" directory not found in sysvol')
        self.assertIn('Policies', ls,
                      msg='"Policies" directory not found in sysvol')
        self.assertNotIn('..', ls,
                         msg='Parent (..) found in directory listing')
        self.assertNotIn('.', ls,
                         msg='Current dir (.) found in directory listing')

        # using a '*' mask should be the same as using no mask
        ls_wildcard = [f['name'] for f in self.smb_conn.list(domain_dir, "*")]
        self.assertEqual(ls, ls_wildcard)

        # applying a mask should only return items that match that mask
        ls_pol = [f['name'] for f in self.smb_conn.list(domain_dir, "Pol*")]
        expected = ["Policies"]
        self.assertEqual(ls_pol, expected)

        # each item in the listing is a has with expected keys
        expected_keys = ['attrib', 'mtime', 'name', 'short_name', 'size']
        for item in self.smb_conn.list(domain_dir):
            for key in expected_keys:
                self.assertIn(key, item,
                              msg="Key '%s' not in listing '%s'" % (key, item))

    def test_deltree(self):
        """The smb.deltree API should delete files and sub-dirs"""
        # create some test sub-dirs
        dirpaths = []
        empty_dirs = []
        cur_dir = test_dir

        for subdir in ["subdir-X", "subdir-Y", "subdir-Z"]:
            path = self.make_sysvol_path(cur_dir, subdir)
            self.smb_conn.mkdir(path)
            dirpaths.append(path)
            cur_dir = path

            # create another empty dir just for kicks
            path = self.make_sysvol_path(cur_dir, "another")
            self.smb_conn.mkdir(path)
            empty_dirs.append(path)

        # create some files in these directories
        filepaths = []
        for subdir in dirpaths:
            for i in range(1, 4):
                contents = "I'm file {0} in dir {1}!".format(i, subdir)
                path = self.make_sysvol_path(subdir, "file-{0}.txt".format(i))
                self.smb_conn.savefile(path, contents.encode('utf8'))
                filepaths.append(path)

        # sanity-check these dirs/files exist
        for subdir in dirpaths + empty_dirs:
            self.assertTrue(self.smb_conn.chkpath(subdir),
                            "Failed to create {0}".format(subdir))
        for path in filepaths:
            self.assertTrue(self.file_exists(path),
                            "Failed to create {0}".format(path))

        # try using deltree to remove a single empty directory
        path = empty_dirs.pop(0)
        self.smb_conn.deltree(path)
        self.assertFalse(self.smb_conn.chkpath(path),
                         "Failed to delete {0}".format(path))

        # try using deltree to remove a single file
        path = filepaths.pop(0)
        self.smb_conn.deltree(path)
        self.assertFalse(self.file_exists(path),
                         "Failed to delete {0}".format(path))

        # delete the top-level dir
        self.smb_conn.deltree(test_dir)

        # now check that all the dirs/files are no longer there
        for subdir in dirpaths + empty_dirs:
            self.assertFalse(self.smb_conn.chkpath(subdir),
                             "Failed to delete {0}".format(subdir))
        for path in filepaths:
            self.assertFalse(self.file_exists(path),
                             "Failed to delete {0}".format(path))

    def file_exists(self, filepath):
        """Returns whether a regular file exists (by trying to open it)"""
        try:
            self.smb_conn.loadfile(filepath)
            exists = True
        except NTSTATUSError as err:
            if (err.args[0] == NT_STATUS_OBJECT_NAME_NOT_FOUND or
                err.args[0] == NT_STATUS_OBJECT_PATH_NOT_FOUND):
                exists = False
            else:
                raise err
        return exists

    def test_unlink(self):
        """
        The smb.unlink API should delete file
        """
        # create the test file
        self.assertFalse(self.file_exists(test_file))
        self.smb_conn.savefile(test_file, binary_contents)
        self.assertTrue(self.file_exists(test_file))

        # delete it and check that it's gone
        self.smb_conn.unlink(test_file)
        self.assertFalse(self.file_exists(test_file))

    def test_chkpath(self):
        """Tests .chkpath determines whether or not a directory exists"""

        self.assertTrue(self.smb_conn.chkpath(test_dir))

        # should return False for a non-existent directory
        bad_dir = self.make_sysvol_path(test_dir, 'dont_exist')
        self.assertFalse(self.smb_conn.chkpath(bad_dir))

        # should return False for files (because they're not directories)
        self.smb_conn.savefile(test_file, binary_contents)
        self.assertFalse(self.smb_conn.chkpath(test_file))

        # check correct result after creating and then deleting a new dir
        new_dir = self.make_sysvol_path(test_dir, 'test-new')
        self.smb_conn.mkdir(new_dir)
        self.assertTrue(self.smb_conn.chkpath(new_dir))
        self.smb_conn.rmdir(new_dir)
        self.assertFalse(self.smb_conn.chkpath(new_dir))

    def test_save_load_text(self):

        self.smb_conn.savefile(test_file, test_contents.encode('utf8'))

        contents = self.smb_conn.loadfile(test_file)
        self.assertEqual(contents.decode('utf8'), test_contents,
                          msg='contents of test file did not match what was written')

        # check we can overwrite the file with new contents
        new_contents = 'wxyz' * 128
        self.smb_conn.savefile(test_file, new_contents.encode('utf8'))
        contents = self.smb_conn.loadfile(test_file)
        self.assertEqual(contents.decode('utf8'), new_contents,
                          msg='contents of test file did not match what was written')

    # this will save/load bytes type
    def test_save_load_string_bytes(self):
        self.smb_conn.savefile(test_file, test_literal_bytes_embed_nulls)

        contents = self.smb_conn.loadfile(test_file)
        self.assertEqual(contents, test_literal_bytes_embed_nulls,
                          msg='contents of test file did not match what was written')

    # this will save/load unicode
    def test_save_load_utfcontents(self):
        self.smb_conn.savefile(test_file, utf_contents.encode('utf8'))

        contents = self.smb_conn.loadfile(test_file)
        self.assertEqual(contents.decode('utf8'), utf_contents,
                          msg='contents of test file did not match what was written')

    # this will save/load bytes type
    def test_save_binary_contents(self):
        self.smb_conn.savefile(test_file, binary_contents)

        contents = self.smb_conn.loadfile(test_file)
        self.assertEqual(contents, binary_contents,
                          msg='contents of test file did not match what was written')

    def make_sysvol_path(self, dirpath, filename):
        # return the dir + filename as a sysvol path
        return os.path.join(dirpath, filename).replace('/', '\\')



# Unix SMB/CIFS implementation.
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2015
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for integer handling in PIDL generated bindings samba.dcerpc.*"""

from samba.dcerpc import server_id, misc, srvsvc, samr
import samba.tests


class IntegerTests(samba.tests.TestCase):

    def test_uint32_into_hyper(self):
        s = server_id.server_id()
        s.unique_id = server_id.NONCLUSTER_VNN
        self.assertEqual(s.unique_id, 0xFFFFFFFF)

    def test_int_into_hyper(self):
        s = server_id.server_id()
        s.unique_id = 1
        self.assertEqual(s.unique_id, 1)

    def test_negative_int_into_hyper(self):
        s = server_id.server_id()

        def assign():
            s.unique_id = -1
        self.assertRaises(OverflowError, assign)

    def test_hyper_into_uint32(self):
        s = server_id.server_id()

        def assign():
            s.vnn = server_id.SERVERID_UNIQUE_ID_NOT_TO_VERIFY
        self.assertRaises(OverflowError, assign)

    def test_hyper_into_int32(self):
        s = srvsvc.NetRemoteTODInfo()

        def assign():
            s.timezone = server_id.SERVERID_UNIQUE_ID_NOT_TO_VERIFY
        self.assertRaises(OverflowError, assign)

    def test_int_into_int32(self):
        s = srvsvc.NetRemoteTODInfo()
        s.timezone = 5
        self.assertEqual(s.timezone, 5)

    def test_uint32_into_int32(self):
        s = srvsvc.NetRemoteTODInfo()

        def assign():
            s.timezone = server_id.NONCLUSTER_VNN
        self.assertRaises(OverflowError, assign)

    def test_long_into_int32(self):
        s = srvsvc.NetRemoteTODInfo()
        # here we force python2 to convert its 32/64 bit python int into
        # an arbitrarily long python long, then reduce the number back
        # down to something that would fit in an int anyway. In a pure
        # python2 world, you could achieve the same thing by writing
        #    s.timezone = 5L
        # but that is a syntax error in py3.
        s.timezone = (5 << 65) >> 65
        self.assertEqual(s.timezone, 5)

    def test_larger_long_int_into_int32(self):
        s = srvsvc.NetRemoteTODInfo()

        def assign():
            s.timezone = 2147483648
        self.assertRaises(OverflowError, assign)

    def test_larger_int_into_int32(self):
        s = srvsvc.NetRemoteTODInfo()
        s.timezone = 2147483647
        self.assertEqual(s.timezone, 2147483647)

    def test_float_into_int32(self):
        s = srvsvc.NetRemoteTODInfo()

        def assign():
            s.timezone = 2.5
        self.assertRaises(TypeError, assign)

    def test_int_float_into_int32(self):
        s = srvsvc.NetRemoteTODInfo()

        def assign():
            s.timezone = 2.0
        self.assertRaises(TypeError, assign)

    def test_negative_int_into_int32(self):
        s = srvsvc.NetRemoteTODInfo()
        s.timezone = -2147483648
        self.assertEqual(s.timezone, -2147483648)

    def test_negative_into_uint32(self):
        s = server_id.server_id()

        def assign():
            s.vnn = -1
        self.assertRaises(OverflowError, assign)

    def test_hyper_into_uint16(self):
        g = misc.GUID()

        def assign():
            g.time_mid = server_id.SERVERID_UNIQUE_ID_NOT_TO_VERIFY
        self.assertRaises(OverflowError, assign)

    def test_int_into_uint16(self):
        g = misc.GUID()

        def assign():
            g.time_mid = 200000
        self.assertRaises(OverflowError, assign)

    def test_negative_int_into_uint16(self):
        g = misc.GUID()

        def assign():
            g.time_mid = -2
        self.assertRaises(OverflowError, assign)

    def test_enum_into_uint16(self):
        g = misc.GUID()
        g.time_mid = misc.SEC_CHAN_DOMAIN
        self.assertEqual(g.time_mid, misc.SEC_CHAN_DOMAIN)

    def test_bitmap_into_uint16(self):
        g = misc.GUID()
        g.time_mid = misc.SV_TYPE_WFW
        self.assertEqual(g.time_mid, misc.SV_TYPE_WFW)

    def test_overflow_bitmap_into_uint16(self):
        g = misc.GUID()

        def assign():
            g.time_mid = misc.SV_TYPE_LOCAL_LIST_ONLY
        self.assertRaises(OverflowError, assign)

    def test_overflow_bitmap_into_uint16_2(self):
        g = misc.GUID()

        def assign():
            g.time_mid = misc.SV_TYPE_DOMAIN_ENUM
        self.assertRaises(OverflowError, assign)

    def test_hyper_into_int64(self):
        s = samr.DomInfo1()

        def assign():
            s.max_password_age = server_id.SERVERID_UNIQUE_ID_NOT_TO_VERIFY
        self.assertRaises(OverflowError, assign)

    def test_int_into_int64(self):
        s = samr.DomInfo1()
        s.max_password_age = 5
        self.assertEqual(s.max_password_age, 5)

    def test_negative_int_into_int64(self):
        s = samr.DomInfo1()
        s.max_password_age = -5
        self.assertEqual(s.max_password_age, -5)

    def test_larger_int_into_int64(self):
        s = samr.DomInfo1()
        s.max_password_age = server_id.NONCLUSTER_VNN
        self.assertEqual(s.max_password_age, 0xFFFFFFFF)

    def test_larger_negative_int_into_int64(self):
        s = samr.DomInfo1()
        s.max_password_age = -2147483649
        self.assertEqual(s.max_password_age, -2147483649)

    def test_int_list_over_list(self):
        g = misc.GUID()
        g.node = [5, 0, 5, 0, 7, 4]
        self.assertEqual(g.node[0], 5)

    def test_long_int_list_over_uint8_list(self):
        g = misc.GUID()
        g.node = [5, 0, 5, 0, 7, 4]
        self.assertEqual(g.node[0], 5)

    def test_negative_list_over_uint8_list(self):
        g = misc.GUID()

        def assign():
            g.node = [-1, 0, 5, 0, 7, 4]
        self.assertRaises(OverflowError, assign)

    def test_overflow_list_over_uint8_list(self):
        g = misc.GUID()

        def assign():
            g.node = [256, 0, 5, 0, 7, 4]
        self.assertRaises(OverflowError, assign)

    def test_short_list_over_uint8_list(self):
        g = misc.GUID()

        def assign():
            g.node = [5, 0, 5]
        self.assertRaises(TypeError, assign)

    def test_long_list_over_uint8_list(self):
        g = misc.GUID()

        def assign():
            g.node = [5, 0, 5, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF]
        self.assertRaises(TypeError, assign)

    # Due to our PIDL bindings generating a python List, modifications
    # to a list of non-objects are not reflected in the C list
    # (modifications objects in lists of objects work because the
    # objects are modified), so changes essentially vanish and are not
    # type checked either.
    def test_assign_into_uint8_list(self):
        g = misc.GUID()
        g.node[1] = 5
        self.assertEqual(g.node[1], 5)

    def test_negative_into_uint8_list(self):
        g = misc.GUID()

        def assign():
            g.node[1] = -1
        self.assertRaises(OverflowError, assign)

    def test_overflow_into_uint8_list(self):
        g = misc.GUID()

        def assign():
            g.node[1] = 256
        self.assertRaises(OverflowError, assign)



# Unix SMB/CIFS implementation. Tests for shares
# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2009
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for samba.hostconfig."""

from samba.hostconfig import SharesContainer
from samba.tests import TestCase


class MockService(object):

    def __init__(self, data):
        self.data = data

    def __getitem__(self, name):
        return self.data[name]


class MockLoadParm(object):

    def __init__(self, data):
        self.data = data

    def __getitem__(self, name):
        return MockService(self.data[name])

    def __len__(self):
        return len(self.data)

    def services(self):
        return self.data.keys()


class ShareTests(TestCase):

    def _get_shares(self, conf):
        return SharesContainer(MockLoadParm(conf))

    def test_len_no_global(self):
        shares = self._get_shares({})
        self.assertEqual(0, len(shares))

    def test_iter(self):
        self.assertEqual([], list(self._get_shares({})))
        self.assertEqual([], list(self._get_shares({"global": {}})))
        self.assertEqual(
            ["bla"],
            list(self._get_shares({"global": {}, "bla": {}})))

    def test_len(self):
        shares = self._get_shares({"global": {}})
        self.assertEqual(0, len(shares))

    def test_getitem_nonexistent(self):
        shares = self._get_shares({"global": {}})
        self.assertRaises(KeyError, shares.__getitem__, "bla")

    def test_getitem_global(self):
        shares = self._get_shares({"global": {}})
        self.assertRaises(KeyError, shares.__getitem__, "global")



# user management
#
# user getpassword command
#
# Copyright Jelmer Vernooij 2010 <jelmer@samba.org>
# Copyright Theresa Halloran 2011 <theresahalloran@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ldb
import samba.getopt as options
from samba.netcmd import CommandError, Option

from .common import (
    GetPasswordCommand,
    gpg_decrypt,
    decrypt_samba_gpg_help,
    virtual_attributes_help
)


class cmd_user_getpassword(GetPasswordCommand):
    """Get the password fields of a user/computer account.

This command gets the logon password for a user/computer account.

The username specified on the command is the sAMAccountName.
The username may also be specified using the --filter option.

The command must be run from the root user id or another authorized user id.
The '-H' or '--URL' option normally only supports ldapi:// or [tdb://] and
can be used to adjust the local path. By default, tdb:// is used.
if the target account is a group managed service account, then in this
case the -H can point to a remote AD DC LDAP server.

The '--attributes' parameter takes a comma separated list of attributes,
which will be printed or given to the script specified by '--script'. If a
specified attribute is not available on an object it's silently omitted.
All attributes defined in the schema (e.g. the unicodePwd attribute holds
the NTHASH) and the following virtual attributes are possible (see --help
for which virtual attributes are supported in your environment):

   virtualClearTextUTF16: The raw cleartext as stored in the
                          'Primary:CLEARTEXT' (or 'Primary:SambaGPG'
                          with '--decrypt-samba-gpg') buffer inside the
                          supplementalCredentials attribute. This typically
                          contains valid UTF-16-LE, but may contain random
                          bytes, e.g. for computer and gMSA accounts.
                          When the account is a group managed service account,
                          and the user is permitted to access
                          msDS-ManagedPassword then the current and previous
                          password can be read over LDAP. Add ;previous=1
                          to read the previous password.

   virtualClearTextUTF8:  As virtualClearTextUTF16, but converted to UTF-8
                          (invalid UTF-16-LE is mapped in the same way as
                          Windows).

   virtualSSHA:           As virtualClearTextUTF8, but a salted SHA-1
                          checksum, useful for OpenLDAP's '{SSHA}' algorithm.

   virtualCryptSHA256:    As virtualClearTextUTF8, but a salted SHA256
                          checksum, useful for OpenLDAP's '{CRYPT}' algorithm,
                          with a $5$... salt, see crypt(3) on modern systems.
                          The number of rounds used to calculate the hash can
                          also be specified. By appending ";rounds=x" to the
                          attribute name i.e. virtualCryptSHA256;rounds=10000
                          will calculate a SHA256 hash with 10,000 rounds.
                          Non-numeric values for rounds are silently ignored.
                          The value is calculated as follows:
                          1) If a value exists in 'Primary:userPassword' with
                             the specified number of rounds it is returned.
                          2) If 'Primary:CLEARTEXT', or 'Primary:SambaGPG'
                             with '--decrypt-samba-gpg'. Calculate a hash with
                             the specified number of rounds.
                          3) Return the first CryptSHA256 value in
                             'Primary:userPassword'.


   virtualCryptSHA512:    As virtualClearTextUTF8, but a salted SHA512
                          checksum, useful for OpenLDAP's '{CRYPT}' algorithm,
                          with a $6$... salt, see crypt(3) on modern systems.
                          The number of rounds used to calculate the hash can
                          also be specified. By appending ";rounds=x" to the
                          attribute name i.e. virtualCryptSHA512;rounds=10000
                          will calculate a SHA512 hash with 10,000 rounds.
                          Non-numeric values for rounds are silently ignored.
                          The value is calculated as follows:
                          1) If a value exists in 'Primary:userPassword' with
                             the specified number of rounds it is returned.
                          2) If 'Primary:CLEARTEXT', or 'Primary:SambaGPG'
                             with '--decrypt-samba-gpg'. Calculate a hash with
                             the specified number of rounds.
                          3) Return the first CryptSHA512 value in
                             'Primary:userPassword'.

   virtualWDigestNN:      The individual hash values stored in
                          'Primary:WDigest' where NN is the hash number in
                          the range 01 to 29.
                          NOTE: As at 22-05-2017 the documentation:
                          3.1.1.8.11.3.1 WDIGEST_CREDENTIALS Construction
                        https://msdn.microsoft.com/en-us/library/cc245680.aspx
                          is incorrect.

   virtualKerberosSalt:   This results the salt string that is used to compute
                          Kerberos keys from a UTF-8 cleartext password.

   virtualSambaGPG:       The raw cleartext as stored in the
                          'Primary:SambaGPG' buffer inside the
                          supplementalCredentials attribute.
                          See the 'password hash gpg key ids' option in
                          smb.conf.

The '--decrypt-samba-gpg' option triggers decryption of the
Primary:SambaGPG buffer. Check with '--help' if this feature is available
in your environment or not (the python-gpgme package is required).  Please
note that you might need to set the GNUPGHOME environment variable.  If the
decryption key has a passphrase you have to make sure that the GPG_AGENT_INFO
environment variable has been set correctly and the passphrase is already
known by the gpg-agent.

Attributes with time values can take an additional format specifier, which
converts the time value into the requested format. The format can be specified
by adding ";format=formatSpecifier" to the requested attribute name, whereby
"formatSpecifier" must be a valid specifier. The syntax looks like:

  --attributes=attributeName;format=formatSpecifier

The following format specifiers are available:
  - GeneralizedTime (e.g. 20210224113259.0Z)
  - UnixTime        (e.g. 1614166392)
  - TimeSpec        (e.g. 161416639.267546892)

Attributes with an original NTTIME value of 0 and 9223372036854775807 are
treated as non-existing value.

Example1:
samba-tool user getpassword TestUser1 --attributes=pwdLastSet,virtualClearTextUTF8

Example2:
samba-tool user getpassword --filter=samaccountname=TestUser3 --attributes=msDS-KeyVersionNumber,unicodePwd,virtualClearTextUTF16

"""

    synopsis = "%prog (<username>|--filter <filter>) [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "versionopts": options.VersionOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--filter", help="LDAP Filter to get password for (must match single account)", type=str),
        Option("--attributes", type=str,
               help=virtual_attributes_help,
               metavar="ATTRIBUTELIST", dest="attributes"),
        Option("--decrypt-samba-gpg",
               help=decrypt_samba_gpg_help,
               action="store_true", default=False, dest="decrypt_samba_gpg"),
    ]

    takes_args = ["username?"]

    def run(self, username=None, H=None, filter=None,
            attributes=None, decrypt_samba_gpg=None,
            sambaopts=None, versionopts=None, hostopts=None,
            credopts=None):
        self.lp = sambaopts.get_loadparm()

        if decrypt_samba_gpg and not gpg_decrypt:
            raise CommandError(decrypt_samba_gpg_help)

        if filter is None and username is None:
            raise CommandError("Either the username or '--filter' must be specified!")

        if filter is None:
            filter = "(&(objectClass=user)(sAMAccountName=%s))" % (ldb.binary_encode(username))

        if attributes is None:
            raise CommandError("Please specify --attributes")

        password_attrs = self.parse_attributes(attributes)

        creds = credopts.get_credentials(self.lp)
        samdb = self.connect_for_passwords(url=hostopts.H, require_ldapi=False, creds=creds)

        obj = self.get_account_attributes(samdb, username,
                                          basedn=None,
                                          filter=filter,
                                          scope=ldb.SCOPE_SUBTREE,
                                          attrs=password_attrs,
                                          decrypt=decrypt_samba_gpg)

        ldif = samdb.write_ldif(obj, ldb.CHANGETYPE_NONE)
        self.outf.write("%s" % ldif)
        self.errf.write("Any available password returned OK\n")



# Create DisplaySpecifiers LDIF (as a string) from the documents provided by
# Microsoft under the WSPP.
#
# Copyright (C) Andrew Kroeger <andrew@id10ts.net> 2009
#
# Based on ms_schema.py
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re


def __read_folded_line(f, buffer):
    """Read a line from an LDIF file, unfolding it"""
    line = buffer

    while True:
        l = f.readline()

        if l[:1] == " ":
            # continued line

            # cannot fold an empty line
            assert(line != "" and line != "\n")

            # preserves '\n '
            line = line + l
        else:
            # non-continued line
            if line == "":
                line = l

                if l == "":
                    # eof, definitely won't be folded
                    break
            else:
                # marks end of a folded line
                # line contains the now unfolded line
                # buffer contains the start of the next possibly folded line
                buffer = l
                break

    return (line, buffer)


# Only compile regexp once.
# Will not match options after the attribute type.
attr_type_re = re.compile("^([A-Za-z][A-Za-z0-9-]*):")


def __read_raw_entries(f):
    """Read an LDIF entry, only unfolding lines"""

    buffer = ""

    while True:
        entry = []

        while True:
            (l, buffer) = __read_folded_line(f, buffer)

            if l[:1] == "#":
                continue

            if l == "\n" or l == "":
                break

            m = attr_type_re.match(l)

            if m:
                if l[-1:] == "\n":
                    l = l[:-1]

                entry.append(l)
            else:
                print("Invalid line: %s" % l, end=' ', file=sys.stderr)
                sys.exit(1)

        if len(entry):
            yield entry

        if l == "":
            break


def fix_dn(dn):
    """Fix a string DN to use ${CONFIGDN}"""

    if dn.find("<Configuration NC Distinguished Name>") != -1:
        dn = dn.replace("\n ", "")
        return dn.replace("<Configuration NC Distinguished Name>", "${CONFIGDN}")
    else:
        return dn


def __write_ldif_one(entry):
    """Write out entry as LDIF"""
    out = []

    for l in entry:
        if l[2] == 0:
            out.append("%s: %s" % (l[0], l[1]))
        else:
            # This is a base64-encoded value
            out.append("%s:: %s" % (l[0], l[1]))

    return "\n".join(out)


def __transform_entry(entry):
    """Perform required transformations to the Microsoft-provided LDIF"""

    temp_entry = []

    for l in entry:
        t = []

        if l.find("::") != -1:
            # This is a base64-encoded value
            t = l.split(":: ", 1)
            t.append(1)
        else:
            t = l.split(": ", 1)
            t.append(0)

        key = t[0].lower()

        if key == "changetype":
            continue

        if key == "distinguishedname":
            continue

        if key == "instancetype":
            continue

        if key == "name":
            continue

        if key == "cn":
            continue

        if key == "objectcategory":
            continue

        if key == "showinadvancedviewonly":
            value = t[1].upper().lstrip().rstrip()
            if value == "TRUE":
                # Remove showInAdvancedViewOnly attribute if it is set to the
                # default value of TRUE
                continue

        t[1] = fix_dn(t[1])

        temp_entry.append(t)

    entry = temp_entry

    return entry


def read_ms_ldif(filename):
    """Read and transform Microsoft-provided LDIF file."""

    out = []

    from io import open
    with open(filename, "r", encoding='latin-1') as f:
        for entry in __read_raw_entries(f):
            out.append(__write_ldif_one(__transform_entry(entry)))

    return "\n\n".join(out) + "\n\n"


if __name__ == '__main__':
    import sys

    try:
        display_specifiers_file = sys.argv[1]
    except IndexError:
        print("Usage: %s display-specifiers-ldif-file.txt" % (sys.argv[0]), file=sys.stderr)
        sys.exit(1)

    print(read_ms_ldif(display_specifiers_file))



# Unix SMB/CIFS implementation.
#
# Model fields.
#
# Copyright (C) Catalyst.Net Ltd. 2023
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import io
from abc import ABCMeta, abstractmethod
from datetime import datetime, timezone
from enum import IntEnum, IntFlag
from xml.etree import ElementTree

from ldb import Dn, MessageElement, binary_encode, string_to_time, timestring
from samba.key_credential_link import KeyCredentialLinkDn
from samba.dcerpc import security
from samba.dcerpc.misc import GUID
from samba.ndr import ndr_pack, ndr_unpack
from samba.nt_time import datetime_from_nt_time, nt_time_from_datetime


class Field(metaclass=ABCMeta):
    """Base class for all fields.

    Each field will need to implement from_db_value and to_db_value.

    A field must correctly support converting both single valued fields,
    and list type fields.

    The only thing many=True does is say the field "prefers" to be a list,
    but really any field can be a list or single value.
    """

    def __init__(self, name, many=False, default=None, hidden=False,
                 readonly=False):
        """Creates a new field, should be subclassed.

        :param name: SamDB field name.
        :param many: If true always convert field to a list when loaded.
        :param default: Default value or callback method (obj is first argument)
        :param hidden: If this is True, exclude the field when calling as_dict()
        :param readonly: If true don't write this value when calling save.
        """
        self.name = name
        self.many = many
        self.hidden = hidden
        self.readonly = readonly

        # This ensures that fields with many=True are always lists.
        # If this is inconsistent anywhere, it isn't so great to use.
        if self.many and default is None:
            self.default = lambda x: list()
        else:
            self.default = default

    @abstractmethod
    def from_db_value(self, samdb, value):
        """Converts value read from the database to Python value.

        :param samdb: SamDB connection
        :param value: MessageElement value from the database
        :returns: Parsed value as Python type
        """
        pass

    @abstractmethod
    def to_db_value(self, samdb, value, flags):
        """Converts value to database value.

        This should return a MessageElement or None, where None means
        the field will be unset on the next save.

        :param samdb: SamDB connection
        :param value: Input value from Python field
        :param flags: MessageElement flags
        :returns: MessageElement or None
        """
        pass

    def expression(self, value):
        """Returns the ldb search expression for this field."""
        return f"({self.name}={binary_encode(value)})"


class IntegerField(Field):
    """A simple integer field, can be an int or list of int."""

    def from_db_value(self, samdb, value):
        """Convert MessageElement to int or list of int."""
        if value is None:
            return
        elif len(value) > 1 or self.many:
            return [int(item) for item in value]
        else:
            return int(value[0])

    def to_db_value(self, samdb, value, flags):
        """Convert int or list of int to MessageElement."""
        if value is None:
            return
        elif isinstance(value, list):
            return MessageElement(
                [str(item) for item in value], flags, self.name)
        else:
            return MessageElement(str(value), flags, self.name)


class BinaryField(Field):
    """Similar to StringField but using bytes instead of str.

    This tends to be quite easy because a MessageElement already uses bytes.
    """

    def from_db_value(self, samdb, value):
        """Convert MessageElement to bytes or list of bytes.

        The values on the MessageElement should already be bytes so the
        cast to bytes() is likely not needed in from_db_value.
        """
        if value is None:
            return
        elif len(value) > 1 or self.many:
            return [bytes(item) for item in value]
        else:
            return bytes(value[0])

    def to_db_value(self, samdb, value, flags):
        """Convert bytes or list of bytes to MessageElement."""
        if value is None:
            return
        elif isinstance(value, list):
            return MessageElement(
                [bytes(item) for item in value], flags, self.name)
        else:
            return MessageElement(bytes(value), flags, self.name)


class StringField(Field):
    """A simple string field, may contain str or list of str."""

    def from_db_value(self, samdb, value):
        """Convert MessageElement to str or list of str."""
        if value is None:
            return
        elif len(value) > 1 or self.many:
            return [str(item) for item in value]
        else:
            return str(value)

    def to_db_value(self, samdb, value, flags):
        """Convert str or list of str to MessageElement."""
        if value is None:
            return
        elif isinstance(value, list):
            return MessageElement(
                [str(item) for item in value], flags, self.name)
        else:
            return MessageElement(str(value), flags, self.name)


class EnumField(Field):
    """A field based around Python's Enum type."""

    def __init__(self, name, enum, many=False, default=None):
        """Create a new EnumField for the given enum class."""
        self.enum = enum
        super().__init__(name, many, default)

    def enum_from_value(self, value):
        """Return Enum instance from value.

        Has a special case for IntEnum as the constructor only accepts int.
        """
        if issubclass(self.enum, (IntEnum, IntFlag)):
            return self.enum(int(str(value)))
        else:
            return self.enum(str(value))

    def from_db_value(self, samdb, value):
        """Convert MessageElement to enum or list of enum."""
        if value is None:
            return
        elif len(value) > 1 or self.many:
            return [self.enum_from_value(item) for item in value]
        else:
            return self.enum_from_value(value)

    def to_db_value(self, samdb, value, flags):
        """Convert enum or list of enum to MessageElement."""
        if value is None:
            return
        elif isinstance(value, list):
            return MessageElement(
                [str(item.value) for item in value], flags, self.name)
        else:
            return MessageElement(str(value.value), flags, self.name)

    def expression(self, value):
        """Returns the ldb search expression for this field."""
        return f"({self.name}={binary_encode(str(value.value))})"


class DateTimeField(Field):
    """A field for parsing ldb timestamps into Python datetime."""

    def from_db_value(self, samdb, value):
        """Convert MessageElement to datetime or list of datetime."""
        if value is None:
            return
        elif len(value) > 1 or self.many:
            return [datetime.fromtimestamp(string_to_time(str(item)),
                                           tz=timezone.utc) for item in value]
        else:
            return datetime.fromtimestamp(string_to_time(str(value)),
                                          tz=timezone.utc)

    def to_db_value(self, samdb, value, flags):
        """Convert datetime or list of datetime to MessageElement."""
        if value is None:
            return
        elif isinstance(value, list):
            return MessageElement(
                [timestring(int(datetime.timestamp(item))) for item in value],
                flags, self.name)
        else:
            return MessageElement(timestring(int(datetime.timestamp(value))),
                                  flags, self.name)


class NtTimeField(Field):
    """18-digit Active Directory timestamps."""

    def from_db_value(self, samdb, value):
        """Convert MessageElement to datetime or list of datetime."""
        if value is None:
            return
        elif len(value) > 1 or self.many:
            return [datetime_from_nt_time(int(item)) for item in value]
        else:
            return datetime_from_nt_time(int(value[0]))

    def to_db_value(self, samdb, value, flags):
        """Convert datetime or list of datetime to MessageElement."""
        if value is None:
            return
        elif isinstance(value, list):
            return MessageElement(
                [str(nt_time_from_datetime(item)) for item in value],
                flags, self.name)
        else:
            return MessageElement(str(nt_time_from_datetime(value)),
                                  flags, self.name)


class RelatedField(Field):
    """A field that automatically fetches the related objects.

    Use sparingly, can be a little slow. If in doubt just use DnField instead.
    """

    def __init__(self, name, model, many=False, default=None):
        """Create a new RelatedField for the given model."""
        self.model = model
        super().__init__(name, many, default)

    def from_db_value(self, samdb, value):
        """Convert Message element to related object or list of objects.

        Note that fetching related items is not using any sort of lazy
        loading so use this field sparingly.
        """
        if value is None:
            return
        elif len(value) > 1 or self.many:
            return [self.model.get(samdb, dn=Dn(samdb, str(item))) for item in value]
        else:
            return self.model.get(samdb, dn=Dn(samdb, str(value)))

    def to_db_value(self, samdb, value, flags):
        """Convert related object or list of objects to MessageElement."""
        if value is None:
            return
        elif isinstance(value, list):
            return MessageElement(
                [str(item.dn) for item in value], flags, self.name)
        else:
            return MessageElement(str(value.dn), flags, self.name)


class DnField(Field):
    """A Dn field parses the current field into a Dn object."""

    def from_db_value(self, samdb, value):
        """Convert MessageElement to a Dn object or list of Dn objects."""
        if value is None:
            return
        elif isinstance(value, Dn):
            return value
        elif len(value) > 1 or self.many:
            return [Dn(samdb, str(item)) for item in value]
        else:
            return Dn(samdb, str(value))

    def to_db_value(self, samdb, value, flags):
        """Convert Dn object or list of Dn objects into a MessageElement."""
        if value is None:
            return
        elif isinstance(value, list):
            return MessageElement(
                [str(item) for item in value], flags, self.name)
        else:
            return MessageElement(str(value), flags, self.name)


class GUIDField(Field):
    """A GUID field decodes fields containing binary GUIDs."""

    def from_db_value(self, samdb, value):
        """Convert MessageElement with a GUID into a str or list of str."""
        if value is None:
            return
        elif len(value) > 1 or self.many:
            return [str(ndr_unpack(GUID, item)) for item in value]
        else:
            return str(ndr_unpack(GUID, value[0]))

    def to_db_value(self, samdb, value, flags):
        """Convert str with GUID into MessageElement."""
        if value is None:
            return
        elif isinstance(value, list):
            return MessageElement(
                [ndr_pack(GUID(item)) for item in value], flags, self.name)
        else:
            return MessageElement(ndr_pack(GUID(value)), flags, self.name)


class SIDField(Field):
    """A SID field encodes and decodes SID data."""

    def from_db_value(self, samdb, value):
        """Convert MessageElement with a GUID into a str or list of str."""
        if value is None:
            return
        elif len(value) > 1 or self.many:
            return [str(ndr_unpack(security.dom_sid, item)) for item in value]
        else:
            return str(ndr_unpack(security.dom_sid, value[0]))

    def to_db_value(self, samdb, value, flags):
        """Convert str with GUID into MessageElement."""
        if value is None:
            return
        elif isinstance(value, list):
            return MessageElement(
                [ndr_pack(security.dom_sid(item)) for item in value],
                flags, self.name)
        else:
            return MessageElement(ndr_pack(security.dom_sid(value)),
                                  flags, self.name)

    def expression(self, value):
        """Returns the ldb search expression for this field."""
        # NOTE: value can be str or `security.dom_sid` so convert to str first.
        return f"({self.name}={binary_encode(str(value))})"


class SDDLField(Field):
    """A SDDL field encodes and decodes SDDL data."""

    def __init__(self,
                 name,
                 *,
                 many=False,
                 default=None,
                 hidden=False,
                 allow_device_in_sddl=True):
        """Create a new SDDLField."""
        self.allow_device_in_sddl = allow_device_in_sddl
        super().__init__(name, many=many, default=default, hidden=hidden)

    def from_db_value(self, samdb, value):
        if value is None:
            return
        elif len(value) > 1 or self.many:
            return [ndr_unpack(security.descriptor, item) for item in value]
        else:
            return ndr_unpack(security.descriptor, value[0])

    def to_db_value(self, samdb, value, flags):
        if value is None:
            return
        elif isinstance(value, list):
            return MessageElement(
                [self.to_db_value(samdb, item, flags)[0] for item in value],
                flags,
                self.name)
        else:
            domain_sid = security.dom_sid(samdb.get_domain_sid())

            # If this is a SDDL string convert it to a descriptor.
            if isinstance(value, str):
                desc = security.descriptor.from_sddl(
                    value, domain_sid,
                    allow_device_in_sddl=self.allow_device_in_sddl)
            else:
                desc = value

            return MessageElement(ndr_pack(desc), flags, self.name)


class BooleanField(Field):
    """A simple boolean field, can be a bool or list of bool."""

    def from_db_value(self, samdb, value):
        """Convert MessageElement into a bool or list of bool."""
        if value is None:
            return
        elif len(value) > 1 or self.many:
            return [str(item) == "TRUE" for item in value]
        else:
            return str(value) == "TRUE"

    def to_db_value(self, samdb, value, flags):
        """Convert bool or list of bool into a MessageElement."""
        if value is None:
            return
        elif isinstance(value, list):
            return MessageElement(
                [str(bool(item)).upper() for item in value], flags, self.name)
        else:
            return MessageElement(str(bool(value)).upper(), flags, self.name)

    def expression(self, value):
        """Returns the ldb search expression for this field."""
        # BooleanField edge case: query by TRUE works but not by FALSE.
        if value is False:
            return f"(!({self.name}=TRUE))"
        else:
            return super().expression(str(value))


class PossibleClaimValuesField(Field):
    """Field for parsing possible values XML for claim types.

    This field will be represented by a list of dicts as follows:

    [
        {"ValueGUID": <GUID>},
        {"ValueDisplayName: "Display name"},
        {"ValueDescription: "Optional description or None for no description"},
        {"Value": <Value>},
    ]

    Note that the GUID needs to be created client-side when adding entries,
    leaving it as None then saving it doesn't generate the GUID.

    The field itself just converts the XML to list and vice versa, it doesn't
    automatically generate GUIDs for entries, this is entirely up to the caller.
    """

    # Namespaces for PossibleValues xml parsing.
    NAMESPACE = {
        "xsd": "http://www.w3.org/2001/XMLSchema",
        "xsi": "http://www.w3.org/2001/XMLSchema-instance",
        "": "http://schemas.microsoft.com/2010/08/ActiveDirectory/PossibleValues"
    }

    def from_db_value(self, samdb, value):
        """Parse MessageElement with XML to list of dicts."""
        if value is not None:
            root = ElementTree.fromstring(str(value))
            string_list = root.find("StringList", self.NAMESPACE)

            values = []
            for item in string_list.findall("Item", self.NAMESPACE):
                values.append({
                    "ValueGUID": item.find("ValueGUID", self.NAMESPACE).text,
                    "ValueDisplayName": item.find("ValueDisplayName",
                                                  self.NAMESPACE).text,
                    "ValueDescription": item.find("ValueDescription",
                                                  self.NAMESPACE).text,
                    "Value": item.find("Value", self.NAMESPACE).text,
                })

            return values

    def to_db_value(self, samdb, value, flags):
        """Convert list of dicts back to XML as a MessageElement."""
        if value is None:
            return

        # Possible values should always be a list of dict, but for consistency
        # with other fields just wrap a single value into a list and continue.
        if isinstance(value, list):
            possible_values = value
        else:
            possible_values = [value]

        # No point storing XML of an empty list.
        # Return None, the field will be unset on the next save.
        if len(possible_values) == 0:
            return

        # root node
        root = ElementTree.Element("PossibleClaimValues")
        for name, url in self.NAMESPACE.items():
            if name == "":
                root.set("xmlns", url)
            else:
                root.set(f"xmlns:{name}", url)

        # StringList node
        string_list = ElementTree.SubElement(root, "StringList")

        # List of values
        for item_dict in possible_values:
            item = ElementTree.SubElement(string_list, "Item")
            item_guid = ElementTree.SubElement(item, "ValueGUID")
            item_guid.text = item_dict["ValueGUID"]
            item_name = ElementTree.SubElement(item, "ValueDisplayName")
            item_name.text = item_dict["ValueDisplayName"]
            item_desc = ElementTree.SubElement(item, "ValueDescription")
            item_desc.text = item_dict["ValueDescription"]
            item_value = ElementTree.SubElement(item, "Value")
            item_value.text = item_dict["Value"]

        # NOTE: indent was only added in Python 3.9 so can't be used yet.
        # ElementTree.indent(root, space="\t", level=0)

        out = io.BytesIO()
        ElementTree.ElementTree(root).write(out,
                                            encoding="utf-16",
                                            xml_declaration=True,
                                            short_empty_elements=False)

        # Back to str as that is what MessageElement needs.
        return MessageElement(out.getvalue().decode("utf-16"), flags, self.name)


class BaseDsdbDnField(Field):
    """Generic DN + Binary field or DN + String field.

    These have this form:

    B:<hex length>:<binary hex>:<ordinary DN>
    S:<utf8 length>:<utf8 string>:<ordinary DN>

    <hex length> is the length of <binary hex> (in decimal), i.e.
    twice the length of the encoded value.

    Subclasses should set dsdb_dn to a BaseDsdbDn subtype.
    """
    dsdb_dn = NotImplemented

    def from_db_value(self, samdb, value):
        """Convert MessageElement to a Dn object or list of Dn objects."""
        if value is None:
            return
        elif isinstance(value, self.dsdb_dn):
            return value
        elif len(value) > 1 or self.many:
            return [self.dsdb_dn(samdb, str(item)) for item in value]
        else:
            return self.dsdb_dn(samdb, str(value))

    def to_db_value(self, samdb, value, flags):
        """Convert Dn object or list of Dn objects into a MessageElement."""
        if value is None:
            return
        elif isinstance(value, list):
            return MessageElement(
                [str(item) for item in value], flags, self.name)
        else:
            return MessageElement(str(value), flags, self.name)


class KeyCredentialLinkDnField(BaseDsdbDnField):
    dsdb_dn = KeyCredentialLinkDn



#!/usr/local/bin/python3.14
# Tests for NDR packing and unpacking of TPM 2.0 public keys
#
# Copyright (C) Gary Lockyer 2025
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os

sys.path.insert(0, "bin/python")
os.environ["PYTHONUNBUFFERED"] = "1"

from samba.dcerpc import tpm20_rsakey_blob
from samba.ndr import ndr_pack, ndr_unpack
from samba.tests import TestCase


class Tpm20RsaKeyBlobTests(TestCase):
    def test_unpack_empty_key_blob(self):
        """
        ensure that a minimal header only TPM_KEY_BLOB
        can be unpacked, then packed into an identical bytes
        """
        key_blob = bytes.fromhex(
            "50 43 50 4D"                     # Magic value PCPM
            "2E 00 00 00"                     # header length
            "02 00 00 00"                     # type TPM 2.0
            "00 00 00 00"                     # flags
            "00 00 00 00"                     # public_length
            "00 00 00 00"                     # private length
            "00 00 00 00"                     # migration public length
            "00 00 00 00"                     # migration private length
            "00 00 00 00"                     # policy digest list length
            "00 00 00 00"                     # PCR binding length
            "00 00 00 00"                     # PCR digest length
            "00 00 00 00"                     # Encrypted secret length
            "00 00 00 00"                     # TPM 1.2 hostage blob length
            "00 00"                           # PCRA Algorithm Id
            "00 00"                           # size
            "00 01"                           # type
            "00 0B"                           # hash algorithm
            "00 00 00 00"                     # attributes
            "00 00"                           # auth_policy empty
            "00 10"                           # algorithm
            "00 14"                           # scheme
            "00 0B"                           # hash algorithm
            "00 00"                           # key bits
            "00 00 00 00"                     # exponent
            "00 00"                           # public key length
        )
        blob = ndr_unpack(tpm20_rsakey_blob.TPM20_RSAKEY_BLOB, key_blob)

        self.assertEqual(blob.type, 2)
        self.assertEqual(blob.public_key.type, 1)
        packed = ndr_pack(blob)
        self.assertEqual(key_blob, packed)

    def test_unpack_sample_key_blob(self):
        """
        ensure that a sample TPM_KEY_BLOB
        can be unpacked, then packed into an identical bytes
        """
        key_blob = bytes.fromhex(
            "50 43 50 4D"                     # Magic value PCPM
            "2E 00 00 00"                     # header length
            "02 00 00 00"                     # type TPM 2.0
            "00 00 00 00"                     # flags
            "00 00 00 00"                     # public_length
            "00 00 00 00"                     # private length
            "00 00 00 00"                     # migration public length
            "00 00 00 00"                     # migration private length
            "00 00 00 00"                     # policy digest list length
            "00 00 00 00"                     # PCR binding length
            "00 00 00 00"                     # PCR digest length
            "00 00 00 00"                     # Encrypted secret length
            "00 00 00 00"                     # TPM 1.2 hostage blob length
            "00 00"                           # PCRA Algorithm Id
            "18 01"                           # size 280 bytes
            "00 01"                           # type
            "00 0B"                           # hash algorithm
            "00 05 24 72"                     # attributes
            "00 00"                           # auth policy"
            "00 10"                           # algorithm
            "00 14"                           # scheme
            "00 0B"                           # hash algorithm
            "08 00"                           # key bits
            "00 00 00 00"                     # exponent
            "01 00"                           # size 256 bytes
            "9A 9E F6 5D E2 92 D6 D0 E5 B3 C4 35 B1 5B 36 F3"
            "9E 83 7B A9 34 AB D9 67 E1 1C 75 43 E5 B6 48 9B"
            "6E CD 8D FC 30 5F 4C B6 8E A0 69 A4 07 21 E7 D7"
            "A1 74 4A 29 BC C9 5D 78 70 C4 3B E4 20 54 BC D0"
            "AA FF 21 44 54 FC 09 08 2A CC DE 44 68 ED 9F B2"
            "3E F7 ED 82 D7 2D 28 74 42 2A 2F 55 A2 E0 DA 45"
            "F1 08 C0 83 8C 95 81 6D 92 CC A8 5D A4 B8 06 8C"
            "76 F5 68 94 E7 60 E6 F4 EE 40 50 28 6C 82 47 89"
            "07 E7 BC 0D 56 5D DA 86 57 E2 CE D3 19 A1 A2 7F"
            "56 F8 99 8B 4A 71 32 6A 57 3B F9 E5 2D 39 35 6E"
            "13 3E 84 DC 5C 96 E1 75 38 C3 AA 23 5B 68 BE 41"
            "52 49 72 7A F6 2A 8F C5 C5 E0 6C DB 99 D1 A8 84"
            "5F 70 21 87 2E A0 D2 68 D3 76 5C 9E D4 9C B5 E1"
            "72 9D 17 8B DC 11 55 09 90 8D 96 F3 68 34 DD 50"
            "63 AC 4A 74 A7 AF 0D DC 15 06 07 D7 5A B3 86 1A"
            "54 96 E0 FA 66 25 31 F5 B4 C7 97 C7 7C 70 94 E3"
        )

        blob = ndr_unpack(tpm20_rsakey_blob.TPM20_RSAKEY_BLOB, key_blob)

        self.assertEqual(blob.type, 2)
        self.assertEqual(blob.public_key.type, 1)
        packed = ndr_pack(blob)
        self.assertEqual(key_blob, packed)

if __name__ == "__main__":
    import unittest

    unittest.main()



# Unix SMB/CIFS implementation.
# Copyright (C) Catalyst.Net Ltd 2025
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""Functions for processing key_credential_link"""

import base64
from hashlib import sha256
import struct
import time
from typing import Optional, Union, Iterable

from cryptography.hazmat.primitives.serialization import (
    load_der_public_key,
    load_pem_public_key,
    PublicFormat,
    Encoding)

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey

from cryptography.x509 import (
    load_pem_x509_certificate,
    load_der_x509_certificate)


from samba import nttime2unix
from samba.samdb import SamDB, BinaryDn
from samba.ndr import ndr_unpack, ndr_pack
from ldb import Dn
from samba.dcerpc import keycredlink, misc


class KeyCredLinkError(Exception):
    """The key credential link is inconsistent."""
    # For bad values handed in, we use ValueError. For internal bad
    # values, we use this.


def key_usage_string(i):
    # there must be a better way.
    for s in ('KEY_USAGE_NGC', 'KEY_USAGE_FIDO', 'KEY_USAGE_FEK',):
        if i == getattr(keycredlink, s):
            return s
    return "unknown"


def nttime_as_date(nt):
    secs = nttime2unix(nt)
    ts = time.gmtime(secs)
    return time.strftime('%Y-%m-%d %H:%M:%S', ts)


class KeyCredentialLinkDn(BinaryDn):
    """KeyCredentialLink attributes are stored as DN+Binary.

    The binary part is a KEYCREDENTIALLINK_BLOB, which is basically an
    array of KEYCREDENTIALLINK_ENTRY collectively describing a public
    key.

    Usually the DN refers to the object the KeyCredentialLink was
    found on.
    """
    # We make .binary a @property, so that BinaryDn's .parse() and
    # .prefix just work without knowing that assigning to .binary is
    # doing validation checks.
    blob = None

    @property
    def binary(self) -> bytes:
        """The binary is stored as a keycredlink.KEYCREDENTIALLINK_BLOB"""
        if self.blob is None:
            return None
        return ndr_pack(self.blob)

    @binary.setter
    def binary(self, value:bytes):
        try:
            self.blob = ndr_unpack(keycredlink.KEYCREDENTIALLINK_BLOB,
                                   value)
        except Exception as e:
            raise ValueError("Could not parse value as KEYCREDENTIALLINK_BLOB "
                             f" (internal error: {e})")

    def get_entry(self, entry_id):
        if self.blob is None:
            raise KeyCredLinkError("no key material")

        for entry in self.blob.entries:
            if entry.identifier == entry_id:
                return entry.value

        raise KeyCredLinkError(f"Key information entry {entry_id} not found")

    def fingerprint(self) -> str:
        """The SHA256 of the key material in DER encoding, formatted
        as hex pairs separated by colons ("hh:hh:...")"""
        # A competing format is '2048 SHA256:<base64bytes>' (ssh style).

        # This sha256 value should also be stored in the KeyID field.
        data = self.get_entry(keycredlink.KeyMaterial)
        hash = sha256(data).digest()
        # Python 3.8+ will do this with hash.hex(':')
        return ':'.join(f'{_:02X}' for _ in hash)

    def description(self, verbosity=2) -> str:
        """Text describing key credential link characteristics.

        verbosity is adjustable between 1 and 3.
        """
        out = []

        def write(msg, verbose_level=0):
            if verbosity > verbose_level:
                out.append(msg)

        write(f'Link target: {self.dn}', 1)
        write(f'Binary Dn: {self}', 2)
        write(f'Key Credential Link Blob version: {self.blob.version}', 2)
        write(f'Number of key entries:            {self.blob.count}', 1)

        write('Key entries:')
        entries = []
        longest = 0
        for description, verbose_level, fn, attr in [
                ("key material fingerprint", 0,
                 lambda x: ':'.join(f"{_:02X}" for _ in x),
                 'KeyID'),
                ("key parameters fingerprint", 2,
                 lambda x: ':'.join(f"{_:02X}" for _ in x),
                 'KeyHash'),
                ("key usage", 1, key_usage_string, 'KeyUsage'),
                ("Device GUID", 1, misc.GUID, 'DeviceId'),
                ("last logon", 0, nttime_as_date,
                 'KeyApproximateLastLogonTimeStamp'),
                ("creation time", 0, nttime_as_date, 'KeyCreationTime'),
                # for now we are ignoring KeySource and CustomKeyInformation
                # KeyMaterial is decoded separately
                ]:

            if verbosity > 1:
                description = f"{description} ({attr})"

            i = getattr(keycredlink, attr)

            try:
                entry = self.get_entry(i)
                value = fn(entry)
            except KeyCredLinkError:
                value = "not found"

            if verbosity > verbose_level:
                entries.append((description, value))
                longest = max(longest, len(description))

        for desc, val in entries:
            write(f"  {desc + ':':{longest + 1}} {val}")

        data = self.get_entry(keycredlink.KeyMaterial)
        key = get_public_key(data, 'der')

        write("RSA public key properties:", 1)
        write(f"  key size: {key.key_size}", 1)
        write(f"  fingerprint: {self.fingerprint()}", 1)

        return '\n'.join(out)

    def key_material(self) -> bytes:
        return self.get_entry(keycredlink.KeyMaterial)

    def as_pem(self) -> str:
        """Get the key out of the keycredlink blob, and return it in
        PEM format as a string.

        PEM is the ASCII format that starts '-----BEGIN PUBLIC KEY-----'.
        """
        # The key is in DER format in an entry in the blob.
        data = self.key_material()
        key = get_public_key(data, 'der')
        pem =  key.public_bytes(Encoding.PEM,
                                PublicFormat.SubjectPublicKeyInfo)
        return pem.decode()


def get_public_key(data:bytes, encoding:Optional[str] = None) -> RSAPublicKey:
    """decode a key in PEM or DER format.

    If it turns out to be a certificate or something, we try to get
    the public key from that.

    So far only RSA keys are supported.
    """
    if encoding is None:
        if data[:11] == b'-----BEGIN ':
            encoding = 'PEM'
        else:
            encoding = 'DER'

    encoding = encoding.upper()

    # The cryptography module also supports ssh keys, PKCS1, and other
    # formats, as well as non-RSA keys and extracting public keys from
    # private. It might not be wise to tolerate all of this, but we
    # can do it by adding to key_fns and cert_fns here.
    if encoding == 'PEM':
        key_fns = [load_pem_public_key]
        cert_fns = [load_pem_x509_certificate]
    elif encoding == 'DER':
        key_fns = [load_der_public_key]
        cert_fns = [load_der_x509_certificate]
    else:
        raise ValueError(f"Public key encoding '{encoding}' not supported "
                         "(try 'PEM' or 'DER')")

    key = None
    for fn in key_fns:
        try:
            key = fn(data)
            break
        except ValueError:
            continue

    if key is None:
        for fn in cert_fns:
            try:
                cert = fn(data)
                key = cert.public_key()
                break
            except ValueError:
                continue

    if key is None:
        raise ValueError("could not decode public key")

    if not isinstance(key, RSAPublicKey):
        raise ValueError("Currently only RSA Public Keys are supported "
                         f"(not '{key}')")

    return key


def kcl_entry_bytes(entry_type:int, data:bytes) -> bytes:
    """helper to pack key credential link entries"""
    return struct.pack('<HB', len(data), entry_type) + data


def create_key_credential_link(samdb: SamDB,
                               target: Union[str, Dn],
                               data: bytes,
                               encoding: Optional[str] = None,
                               force: bool = False):
    """Convert a public key in a common format into a binary DN"""
    if not force:
        res = samdb.search(base=target)
        if len(res) == 0:
            raise ValueError(f"link target {target} does not exist")

    if encoding == 'auto':
        encoding = None

    key = get_public_key(data, encoding)

    if key.key_size != 2048:
        # According to [MS-ADTS] 2.2.20.5.1, KEY_USAGE_NGC means a
        # 2048 bit public key.
        if not force:
            raise ValueError(f"2048 bit RSA key expected, not {key.key_size}")

    key_bytes = key.public_bytes(Encoding.DER,
                                 PublicFormat.SubjectPublicKeyInfo)

    # that's the key.
    # but there's more.
    kcl_header = bytes.fromhex("00 02 00 00")  # Always version 2

    # Entries are added in the enum order, as follows.
    #
    # Here '**' means MUST exist, '*' means SHOULD, and '-' means
    # SHOULD which we ignore. We ignore all the un-SHOULDed values
    # ([MS-ADTS] 2.2.20.6). For KeyUsage, only use KEY_USAGE_NGC.
    #
    # ** 1 KeyID            hash of the key material
    #  * 2 KeyHash          hash of following entries (i.e. 3, 4, 9)
    # ** 3 KeyMaterial      the key
    # ** 4 KeyUsage         KEY_USAGE_NGC, KEY_USAGE_FIDO, or KEY_USAGE_FEK
    #    5 KeySource        KEY_SOURCE_AD.
    #    6 DeviceId         16 byte device ID (GUID, I guess) or zeros
    #    7 CustomKeyInformation  CUSTOM_KEY_INFORMATION struct
    #  - 8 KeyApproximateLastLogonTimeStamp  nttime
    #  * 9 KeyCreationTime   nttime

    # sha256 of the actual key
    kcl_key_id = kcl_entry_bytes(keycredlink.KeyID,
                                 sha256(key_bytes).digest())

    # the actual key
    kcl_material = kcl_entry_bytes(keycredlink.KeyMaterial,
                                   key_bytes)

    # always KEY_USAGE_NGC
    kcl_key_usage = kcl_entry_bytes(keycredlink.KeyUsage,
                                    keycredlink.KEY_USAGE_NGC.to_bytes(1, byteorder='big'))

    # nttime for now
    kcl_creation = kcl_entry_bytes(keycredlink.KeyCreationTime,
                                   struct.pack('<Q', samdb.get_nttime()))

    # always KEY_SOURCE_AD
    #kcl_key_source = kcl_entry_bytes(keycredlink.KeySource,
    #                                 KEY_SOURCE_AD.to_bytes())

    # the KeyHash field is a sha256 of all the values after the
    # KeyHash field.

    kcl_key_hash = kcl_entry_bytes(keycredlink.KeyHash,
                                   sha256(kcl_material +
                                          kcl_key_usage +
                                          kcl_creation).digest())

    kcl_bytes = (kcl_header +
                 kcl_key_id +
                 kcl_key_hash +
                 kcl_material +
                 kcl_key_usage +
                 kcl_creation)

    k = KeyCredentialLinkDn.from_bytes_and_dn(samdb, kcl_bytes, target)
    return k

def kcl_in_list(kcl: KeyCredentialLinkDn, others: Iterable[KeyCredentialLinkDn]):
    """True if kcl is in the list, otherwise False, disregarding
    everything except key material and DN for the comparison.
    """
    # this helps us avoid duplicate key credential links, which are
    # otherwise disallowed only if all fields are identical, but which
    # are generally useless.
    km = kcl.key_material()
    for other in others:
        if km == other.key_material() and kcl.dn == other.dn:
            return True
    return False


def filter_kcl_list(samdb: SamDB,
                    keycredlinks: Iterable[KeyCredentialLinkDn],
                    link_target: Optional[str] = None,
                    fingerprint: Optional[str] = None) -> list:
    """Select only the input links that match at least one of the
    criteria.
    """
    # used in samba-tool X keytrust delete
    selected = []
    filters = []
    if link_target is not None:
        target_dn = Dn(samdb, link_target)
        filters.append(lambda x: x.dn == target_dn)

    if fingerprint is not None:
        fingerprint = fingerprint.upper()
        filters.append(lambda x: x.fingerprint() == fingerprint)

    for x in keycredlinks:
        for fn in filters:
            if fn(x):
                selected.append(x)
                break

    return selected
